# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified: 2023 James Flynn
# Original:
# https://github.com/deepmind/open_spiel/blob/master/open_spiel/python/algorithms/cfr.py

"""Python implementation of the extensive-form regret minimization algorithm.

See: "Efficient Deviation Types and Learning
      for Hindsight Rationality in Extensive-Form Games",
      Morrill et al. 2021b,
      https://arxiv.org/abs/2102.06973

One iteration of EFR consists of:
1) Compute current strategy from regrets (e.g. using Regret Matching).
2) Compute values using the current strategy
3) Compute regrets from these values

The average policy converges to a Nash Equilibrium
rather than the current policy.
"""

import collections
import copy

import attr
import numpy as np
from scipy import linalg

from open_spiel.python import policy
import pyspiel


@attr.s
class _InfoStateNode(object):
  """An object wrapping values associated to an information state."""

  # The list of the legal actions.
  legal_actions = attr.ib()
  index_in_tabular_policy = attr.ib()
  # The newly availible deviations + the old ones
  relizable_deviations = attr.ib()
  # Player -> state -> action -> prob
  current_history_probs = attr.ib()

  # An array representing the preceeding actions played
  # upto this information state.
  history = attr.ib()

  cumulative_regret = attr.ib(factory=lambda: collections.defaultdict(float))
  # The sum of all prior iteration's policies
  cumulative_policy = attr.ib(factory=lambda: collections.defaultdict(float))

  # A dictionary mapping each deviation to their "y values"
  # for the current iteration.
  y_values = attr.ib(factory=lambda: collections.defaultdict(float))


class _EFRSolverBase(object):
  """The base EFR solver class.

  The main iteration loop is implemented in `evaluate_and_update_policy`:
  ```python
    game = pyspiel.load_game("game_name")
    initial_state = game.new_initial_state()
    solver = Solver(game)
    for i in range(num_iterations):
      solver.evaluate_and_update_policy()
      solver.current_policy()  # Access the current policy
      solver.average_policy()  # Access the average policy
  ```
  """

  def __init__(self, game, deviation_gen):
    """Initializer.

    Args:
      game: The `pyspiel.Game` to run on.
      deviation_gen: a function that accepts (num_actions : int,
                                              history, prior_legal_actions)
        and returns a list containing`LocalDeviationWithTimeSelection` objects
        of the realisable deviations of a described type
        (e.g blind causal deviations) and given the information state described
        by the function parameters.
    """
    # pyformat: enable
    assert game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL, (
        "EFR requires sequential games. If you're trying to run it "
        + "on a simultaneous (or normal-form) game, please first transform it "
        + "using turn_based_simultaneous_game."
    )

    self._game = game
    self._num_players = game.num_players()
    self._root_node = self._game.new_initial_state()

    # This is for returning the current policy and average policy to a caller
    self._current_policy = policy.TabularPolicy(game)
    self._average_policy = self._current_policy.__copy__()
    self._deviation_gen = deviation_gen

    self._info_state_nodes = {}
    hist = {player: [] for player in range(self._num_players)}
    empty_path_indices = [[] for _ in range(self._num_players)]

    self._initialize_info_state_nodes(self._root_node, hist, empty_path_indices)

    self._iteration = 1  # For possible linear-averaging.

  def return_cumulative_regret(self):
    """Returns a dictionary mapping.

    The mapping is from every information state to its associated regret
    (accumulated over all iterations).
    """
    return {
        list(self._info_state_nodes.keys())[i]: list(
            self._info_state_nodes.values()
        )[i].cumulative_regret
        for i in range(len(self._info_state_nodes.keys()))
    }

  def current_policy(self):
    """Returns the current policy as a TabularPolicy.

    WARNING: The same object, updated in-place will be returned! You can copy
    it (or its `action_probability_array` field).

    For EFR, this policy does not necessarily have to converge.
    """
    return self._current_policy

  def average_policy(self):
    """Returns the average of all policies iterated.

    WARNING: The same object, updated in-place will be returned! You can copy it
    (or its `action_probability_array` field).

    This average policy converges to a equilibrium policy as the number
    of iterations increases (equilibrium type depends on learning
    deviations used).

    The policy is computed using the accumulated policy probabilities computed
    using `evaluate_and_update_policy`.

    Returns:
    A `policy.TabularPolicy` object (shared between calls) giving the (linear)
    time averaged policy (weighted by player reach probabilities) for all
    players.
    """
    _update_average_policy(self._average_policy, self._info_state_nodes)
    return self._average_policy

  def _initialize_info_state_nodes(self, state, history, path_indices):
    """Initializes info_state_nodes.

    Create one _InfoStateNode per infoset. We could also initialize the node
    when we try to access it and it does not exist.

    Generates all deviations that are realisable at this state and stores
    the history and preceeding state policy information to create memory states
    and calculate the memory reach probability for each deviation.

    Args:
      state: The current state in the tree traversal. This should be the
        root node when we call this function from the EFR solver.
      history: an arrays of the preceeding actions taken prior to the state
        for each player.
      path_indices: a 3d array [player number]x[preceeding state]x[legal actions
        for state, index of the policy for this state in TabularPolicy].
    """
    if state.is_terminal():
      return

    if state.is_chance_node():
      for action, unused_action_prob in state.chance_outcomes():
        self._initialize_info_state_nodes(
            state.child(action), history, path_indices
        )
      return

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)
    info_state_node = self._info_state_nodes.get(info_state)
    if info_state_node is None:
      legal_actions = state.legal_actions(current_player)
      info_state_node = _InfoStateNode(
          legal_actions=legal_actions,
          index_in_tabular_policy=self._current_policy.state_lookup[info_state],
          relizable_deviations=None,
          history=history[current_player].copy(),
          current_history_probs=copy.deepcopy(path_indices[current_player]),
      )
      prior_possible_actions = []
      for i in range(len(info_state_node.current_history_probs)):
        prior_possible_actions.append(
            info_state_node.current_history_probs[i][0]
        )
      prior_possible_actions.append(info_state_node.legal_actions)

      info_state_node.relizable_deviations = self._deviation_gen(
          len(info_state_node.legal_actions),
          info_state_node.history,
          prior_possible_actions,
      )
      self._info_state_nodes[info_state] = info_state_node

    legal_actions = state.legal_actions(current_player)

    for action in info_state_node.legal_actions:
      new_path_indices = copy.deepcopy(path_indices)
      new_path_indices[current_player].append(
          [legal_actions, info_state_node.index_in_tabular_policy]
      )
      new_history = copy.deepcopy(history)
      new_history[current_player].append(action)
      assert len(new_history[current_player]) == len(
          new_path_indices[current_player]
      )

      self._initialize_info_state_nodes(
          state.child(action), new_history, new_path_indices
      )

  def _update_current_policy(self, state, current_policy):
    """Updated the current policy.

    Updated in order so that memory reach probs are defined wrt to the new
    strategy.

    Note that the function is called recursively (first call should
    be the root).

    Additionally, to update the strategy for a given state we require
    the (t+1)th strategy for all prior states.

    Args:
      state: the state of which to update the strategy.
      current_policy: the (t+1)th strategy that is being recursively computed,
        see the function description for more detail.
    """

    if state.is_terminal():
      return
    elif not state.is_chance_node():
      current_player = state.current_player()
      info_state = state.information_state_string(current_player)
      info_state_node = self._info_state_nodes[info_state]
      deviations = info_state_node.relizable_deviations
      for devation in range(len(deviations)):
        mem_reach_probs = create_probs_from_index(
            info_state_node.current_history_probs, current_policy
        )
        deviation_reach_prob = deviations[
            devation
        ].player_deviation_reach_probability(mem_reach_probs)
        y_increment = (
            max(0, info_state_node.cumulative_regret[devation])
            * deviation_reach_prob
        )
        info_state_node.y_values[deviations[devation]] = (
            info_state_node.y_values[deviations[devation]] + y_increment
        )

      state_policy = current_policy.policy_for_key(info_state)
      for action, value in self._regret_matching(info_state_node).items():
        state_policy[action] = value

      for action in info_state_node.legal_actions:
        new_state = state.child(action)
        self._update_current_policy(new_state, current_policy)
    else:
      for action, _ in state.chance_outcomes():
        new_state = state.child(action)
        self._update_current_policy(new_state, current_policy)

  # Path to state probability ignores chance probabilty as this is stored as
  # new_reach_probabilities[-1]
  def _compute_cumulative_immediate_regret_for_player(
      self, state, policies, reach_probabilities, player
  ):
    """Increments the immediate regrets and policy.

    Increments are done for `player` of all realisable deviations at this state.

    Args:
      state: The initial game state to analyze from.
      policies: A list of `num_players` callables taking as input an
        `info_state_node` and returning a {action: prob} dictionary.
      reach_probabilities: The probability for each player of reaching `state`
        as a numpy array [prob for player 0, for player 1,..., for chance].
        `reach_probabilities[player]` will work in all cases.
      player: The 0-indexed player to update the values for. If `None`, the
        update for all players will be performed.

    Returns:
      The utility of `state` for all players, assuming all players follow the
      current policy defined by `self.Policy`.
    """
    if state.is_terminal():
      return np.asarray(state.returns())

    if state.is_chance_node():
      state_value = 0.0
      for action, action_prob in state.chance_outcomes():
        assert action_prob > 0
        new_state = state.child(action)
        new_reach_probabilities = reach_probabilities.copy()
        new_reach_probabilities[-1] *= action_prob

        state_value += (
            action_prob
            * self._compute_cumulative_immediate_regret_for_player(
                new_state, policies, new_reach_probabilities, player
            )
        )
      return state_value

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)

    # No need to continue on this history branch as no update will be performed
    # for any player.
    # The value we return here is not used in practice. If the conditional
    # statement is True, then the last taken action has probability 0 of
    # occurring, so the returned value is not impacting the parent node value.
    if all(reach_probabilities[:-1] == 0):
      return np.zeros(self._num_players)

    state_value = np.zeros(self._num_players)

    # The utilities of the children states are computed recursively. As the
    # regrets are added to the information state regrets for each state in that
    # information state, the recursive call can only be made once per child
    # state. Therefore, the utilities are cached.
    children_utilities = {}

    info_state_node = self._info_state_nodes[info_state]
    # Reset y values
    info_state_node.y_values = collections.defaultdict(float)
    if policies is None:
      info_state_policy = self._get_infostate_policy(info_state)
    else:
      info_state_policy = policies[current_player](info_state)

    reach_prob = reach_probabilities[current_player]
    for action in state.legal_actions():
      action_prob = info_state_policy.get(action, 0.0)
      info_state_node.cumulative_policy[action] = (
          info_state_node.cumulative_policy[action] + action_prob * reach_prob
      )
      new_state = state.child(action)
      new_reach_probabilities = reach_probabilities.copy()
      assert action_prob <= 1
      new_reach_probabilities[current_player] *= action_prob
      child_utility = self._compute_cumulative_immediate_regret_for_player(
          new_state,
          policies=policies,
          reach_probabilities=new_reach_probabilities,
          player=player,
      )

      state_value += action_prob * child_utility
      children_utilities[action] = child_utility

    counterfactual_reach_prob = np.prod(
        reach_probabilities[:current_player]
    ) * np.prod(reach_probabilities[current_player + 1 :])

    state_value_for_player = state_value[current_player]
    deviations = info_state_node.relizable_deviations
    for deviation_index in range(len(deviations)):
      deviation = deviations[deviation_index]
      deviation_strategy = deviation.deviate(
          strat_dict_to_array(self._get_infostate_policy(info_state))
      )

      player_child_utilities = np.array(list(children_utilities.values()))[
          :, current_player
      ]
      devation_cf_value = np.inner(
          np.transpose(deviation_strategy), player_child_utilities
      )

      memory_reach_probs = create_probs_from_index(
          info_state_node.current_history_probs, self.current_policy()
      )
      player_current_memory_reach_prob = (
          deviation.player_deviation_reach_probability(memory_reach_probs)
      )

      deviation_regret = player_current_memory_reach_prob * (
          (devation_cf_value * counterfactual_reach_prob)
          - (counterfactual_reach_prob * state_value_for_player)
      )

      info_state_node.cumulative_regret[deviation_index] += deviation_regret
    return state_value

  def _get_infostate_policy(self, info_state_str):
    """Returns an {action: prob} dictionary for the policy on `info_state`."""
    info_state_node = self._info_state_nodes[info_state_str]
    prob_vec = self._current_policy.action_probability_array[
        info_state_node.index_in_tabular_policy
    ]
    return {
        action: prob_vec[action] for action in info_state_node.legal_actions
    }


class _EFRSolver(_EFRSolverBase):

  def evaluate_and_update_policy(self):
    """Performs a single step of policy evaluation and policy improvement."""
    self._compute_cumulative_immediate_regret_for_player(
        self._root_node,
        policies=None,
        reach_probabilities=np.ones(self._game.num_players() + 1),
        player=None,
    )
    self._update_current_policy(self._root_node, self._current_policy)
    self._iteration += 1


class EFRSolver(_EFRSolver):
  """Implements the EFR algorithm with several deviation types.

  See: https://arxiv.org/abs/2102.06973
  """

  def __init__(self, game, deviations_name):
    """Initializer.

    Args:
      game: The `pyspiel.Game` to run on.
      deviations_name: the name of the deviation type to use for
        accumulating regrets and calculating the strategy at the next timestep.

    Deviation types implemented are "blind action", "informed action",
    "blind cf", "informed counterfactual", "blind partial sequence",
    "counterfactual partial sequence", "casual partial sequence",
    "twice informed partial sequence", "single target behavioural".

    See "Efficient Deviation Types and Learning for Hindsight Rationality in
    Extensive-Form Games" by D. Morrill et al. 2021b
    for the full definition of each type.
    """

    # external_only = True leads to a shortcut in the computation of the next
    # timesteps strategy from the regrets
    external_only = False
    deviation_sets = None  # pylint: disable=unused-variable

    if deviations_name in {"blind action"}:
      deviation_sets = return_blind_action
      external_only = True
    elif deviations_name in {"informed action"}:
      deviation_sets = return_informed_action
    elif deviations_name in {"blind cf", "blind counterfactual"}:
      deviation_sets = return_blind_cf
      external_only = True
    elif deviations_name in {"informed cf", "informed counterfactual"}:
      deviation_sets = return_informed_cf
    elif deviations_name in {"bps", "blind partial sequence"}:
      deviation_sets = return_blind_partial_sequence
      external_only = True
    elif deviations_name in {
        "cfps",
        "cf partial sequence",
        "counterfactual partial sequence",
    }:
      deviation_sets = return_cf_partial_sequence
    elif deviations_name in {"csps", "casual partial sequence"}:
      deviation_sets = return_cs_partial_sequence
    elif deviations_name in {"tips", "twice informed partial sequence"}:
      deviation_sets = return_twice_informed_partial_sequence
    elif deviations_name in {"bhv", "single target behavioural", "behavioural"}:
      deviation_sets = return_behavourial
    else:
      raise ValueError(
          "Unsupported Deviation Set Passed              As Constructor"
          " Argument"
      )
    super(EFRSolver, self).__init__(game, deviation_sets)
    self._external_only = external_only

  def _regret_matching(self, info_set_node):
    """Returns an info state policy.

    The info state policy returned is the one obtained by applying
    regret-matching function over all deviations and time selection functions.

    Args:
      info_set_node: the info state node to compute the policy for.

    Returns:
      A dict of action -> prob for all legal actions of the
      info_set_node.
    """
    legal_actions = info_set_node.legal_actions
    num_actions = len(legal_actions)
    info_state_policy = None
    z = sum(info_set_node.y_values.values())

    # The fixed point solution can be directly obtained through the
    # weighted regret matrix if only external deviations are used.
    if self._external_only and z > 0:
      weighted_deviation_matrix = np.zeros((num_actions, num_actions))
      for dev in list(info_set_node.y_values.keys()):
        weighted_deviation_matrix += (
            info_set_node.y_values[dev] / z
        ) * dev.return_transform_matrix()
      new_strategy = weighted_deviation_matrix[:, 0]
      info_state_policy = dict(zip(legal_actions, new_strategy))

    # Full regret matching by finding the least squares solution to the
    # fixed point of the EFR regret matching function.
    # Last row of matrix and the column entry minimises the solution
    # towards a strategy.
    elif z > 0:
      weighted_deviation_matrix = -np.eye(num_actions)

      for dev in list(info_set_node.y_values.keys()):
        weighted_deviation_matrix += (
            info_set_node.y_values[dev] / z
        ) * dev.return_transform_matrix()

      normalisation_row = np.ones(num_actions)
      weighted_deviation_matrix = np.vstack(
          [weighted_deviation_matrix, normalisation_row]
      )
      b = np.zeros(num_actions + 1)
      b[num_actions] = 1
      b = np.reshape(b, (num_actions + 1, 1))

      strategy = linalg.lstsq(weighted_deviation_matrix, b)[0]

      # Adopt same clipping strategy as paper author's code.
      np.clip(strategy, a_min=0, a_max=1, out=strategy)
      strategy = strategy / np.sum(strategy)

      info_state_policy = dict(zip(legal_actions, strategy[:, 0]))
    # Use a uniform strategy as sum of all regrets is negative.
    else:
      unif_policy_value = 1.0 / num_actions
      info_state_policy = {
          legal_actions[index]: unif_policy_value
          for index in range(num_actions)
      }
    return info_state_policy


def _update_average_policy(average_policy, info_state_nodes):
  """Updates in place `average_policy` to the average of all policies iterated.

  This function is a module level function to be reused by both CFRSolver and
  CFRBRSolver.

  Args:
    average_policy: A `policy.TabularPolicy` to be updated in-place.
    info_state_nodes: A dictionary {`info_state_str` -> `_InfoStateNode`}.
  """
  for info_state, info_state_node in info_state_nodes.items():
    info_state_policies_sum = info_state_node.cumulative_policy
    state_policy = average_policy.policy_for_key(info_state)
    probabilities_sum = sum(info_state_policies_sum.values())
    if probabilities_sum == 0:
      num_actions = len(info_state_node.legal_actions)
      for action in info_state_node.legal_actions:
        state_policy[action] = 1 / num_actions
    else:
      for action, action_prob_sum in info_state_policies_sum.items():
        state_policy[action] = action_prob_sum / probabilities_sum


def strat_dict_to_array(strategy_dictionary):
  """A helper function to convert the strategy dictionary mapping.

  Conversion applies action -> prob value to an array.

  Args:
    strategy_dictionary: a dictionary action -> prob value.

  Returns:
    strategy_array: an array with the ith action's value at the i-1th index.
  """
  actions = list(strategy_dictionary.keys())
  strategy_array = np.zeros((len(actions), 1))
  for action in range(len(actions)):
    strategy_array[action][0] = strategy_dictionary[actions[action]]
  return strategy_array


def array_to_strat_dict(strategy_array, legal_actions):
  """A helper function to convert a strategy.

  Converts a strategy array to an action -> prob value dictionary.

  Args:
    strategy_array: an array with the ith action's value at the i-1th index.
    legal_actions: the list of all legal actions at the current state.

  Returns:
    strategy_dictionary: a dictionary action -> prob value.
  """
  return dict(zip(legal_actions, strategy_array))


def create_probs_from_index(indices, current_policy):
  path_to_state = []
  if indices is None or not indices:
    return []
  for index in indices:
    strat_dict = array_to_strat_dict(
        current_policy.action_probability_array[index[1]], index[0]
    )
    path_to_state.append(strat_dict)
  return path_to_state


# Deviation set definitions
def return_blind_action(num_actions, history, _):
  """Returns an array of all Blind Action deviations.

  Returns an array of all Blind Action deviations. with respect to an with
  respect to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Blind Action deviations that are realizable at the information set.
  """
  memory_weights = [np.full(len(history), 1)]
  prior_actions_in_memory = history
  return return_all_external_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_informed_action(num_actions, history, _):
  """Returns an array of all Informed Action deviations.

  Returns an array of all Informed Action deviations with respect to an
  information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Informed Action deviations that are realizable at the information set.
  """
  memory_weights = [np.full(len(history), 1)]
  prior_actions_in_memory = history
  return return_all_non_identity_internal_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_blind_cf(num_actions, history, _):
  """Returns an array of all Blind Counterfactual deviations.

  Returns an array of all Blind Counterfactual deviations with respect to an
  information set.

  Note: EFR using only Blind Counterfactual deviations is equivalent
  to vanilla Counterfactual Regret Minimisation (CFR).
  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Blind CF deviations that are realizable at the information set.
  """
  memory_weights = [None]
  prior_actions_in_memory = np.zeros(len(history))
  return return_all_external_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_informed_cf(num_actions, history, _):
  """Returns an array of all Informed Counterfactual deviations.

  Returns an array of all Informed Counterfactual deviations with respect with
  respect to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Informed CF deviations that are realizable at the information set.
  """
  memory_weights = [None]
  prior_actions_in_memory = np.zeros(len(history))
  return return_all_non_identity_internal_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_blind_partial_sequence(num_actions, history, _):
  """Returns an array of all Blind Partial Sequence deviations (BPS).

  Returns an array of all Blind Partial Sequence deviations (BPS) with respect
  to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    BPS deviations that are realizable at the information set.
  """
  prior_actions_in_memory = history
  memory_weights = [None]
  if history:
    memory_weights.append(np.ones(len(history)))
  for i in range(len(history)):
    possible_memory_weight = np.zeros(len(history))
    possible_memory_weight[0:i] = np.full(i, 1.0)
    memory_weights.append(possible_memory_weight)
  return return_all_external_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_cf_partial_sequence(num_actions, history, _):
  """Returns an array of all Counterfactual Partial Sequence deviations (CFPS).

  Returns an array of all Counterfactual Partial Sequence deviations (CFPS)
  with respect to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set.
    history: an array containing the prior actions played by the `player` to
      reach the information set.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent
    all CFPS deviations that are realizable at the information set.
  """
  prior_actions_in_memory = history
  memory_weights = [None]
  if history:
    memory_weights.append(np.ones(len(history)))
  for i in range(len(history)):
    possible_memory_weight = np.zeros(len(history))
    possible_memory_weight[0:i] = np.full(i, 1.0)
    memory_weights.append(possible_memory_weight)
  return return_all_non_identity_internal_deviations(
      num_actions, memory_weights, prior_actions_in_memory
  )


def return_cs_partial_sequence(num_actions, history, prior_legal_actions):
  """Returns an array of all Casual Partial Sequence deviations.

  Returns an array of all Casual Partial Sequence deviations with respect to
  an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set
    history: an array containing the prior actions played by the `player` to
      reach the information set.
    prior_legal_actions: a 2d array containing the legal actions for each
      preceeding state.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Casual Partial Sequence deviations that are realizable at the
    information set.
  """
  prior_actions_in_memory = history
  external_memory_weights = []

  for i in range(len(history)):
    possible_memory_weight = np.zeros(len(history))
    possible_memory_weight[0:i] = np.full(i, 1.0)
    external_memory_weights.append(possible_memory_weight)

  external = return_all_external_modified_deviations(
      num_actions,
      external_memory_weights,
      prior_legal_actions,
      prior_actions_in_memory,
  )
  internal = return_blind_action(num_actions, history, None)

  cf_ext = return_informed_cf(num_actions, history, None)
  cf_int = return_blind_cf(num_actions, history, None)

  return np.concatenate((external, internal, cf_ext, cf_int))


def return_cs_partial_sequence_orginal(
    num_actions, history, prior_legal_actions
):
  """Returns an array of all Casual Partial Sequence deviations.

  Returns an array of all Casual Partial Sequence deviations with respect to
  an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set
    history: an array containing the prior actions played by the `player` to
      reach the information set.
    prior_legal_actions: a 2d array containing the legal actions for each
      preceeding state.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent all
    Casual Partial Sequence deviations that are realizable at the
    information set.
  """
  prior_actions_in_memory = history
  external_memory_weights = []

  for i in range(len(history)):
    possible_memory_weight = np.zeros(len(history))
    possible_memory_weight[0:i] = np.full(i, 1.0)
    external_memory_weights.append(possible_memory_weight)

  external = return_all_external_modified_deviations(
      num_actions,
      external_memory_weights,
      prior_legal_actions,
      prior_actions_in_memory,
  )
  internal = return_informed_action(num_actions, history, None)

  cf_ext = return_informed_cf(num_actions, history, None)
  return np.concatenate((external, internal, cf_ext))


def return_twice_informed_partial_sequence(
    num_actions, history, prior_legal_actions
):
  """Returns an array of all Twice Informed Partial Sequence (TIPS) deviations.

  Returns an array of all Twice Informed Partial Sequence (TIPS) deviations
  with respect to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set
    history: an array containing the prior actions played by the `player` to
      reach the information set.
    prior_legal_actions: a 2d array containing the legal actions for each
      preceeding state.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent
    all TIPS deviations that are realizable at theinformation set.
  """
  prior_actions_in_memory = history
  memory_weights = []

  for i in range(len(history)):
    possible_memory_weight = np.zeros(len(history))
    possible_memory_weight[0:i] = np.full(i, 1.0)
    memory_weights.append(possible_memory_weight)

  internal = return_all_internal_modified_deviations(
      num_actions, memory_weights, prior_legal_actions, prior_actions_in_memory
  )

  cf_int = return_informed_cf(num_actions, history, None)
  return np.concatenate((internal, cf_int))


def generate_all_action_permutations(current_stem, remaining_actions):
  """Return a List of all possible game continuations.

  Return a List of all possible game continuations playing on from the
  current stem and with playing from the set of remaining actions.
  `current_stem` = "" generates all possible playthroughs from the current
  information state.

  Args:
     current_stem: the prior sequence of actions to be completed by the
       remaining actions
     remaining_actions: a 2d array of [subsequent states]x[possible actions]

  Returns:
     An array with each element being the current stem joined with a possible
     permuation of remaining actions
  """
  if not remaining_actions:
    return [np.array(current_stem)]
  else:
    next_actions = remaining_actions[0]
    permutations = []
    for action in next_actions:
      next_stem = current_stem.copy()
      next_stem.append(action)
      next_remaining_actions = remaining_actions[1:]
      prev_permutations = generate_all_action_permutations(
          next_stem, next_remaining_actions
      )
      for i in prev_permutations:
        permutations.append(i)
    return permutations


def return_behavourial(num_actions, history, prior_legal_actions):
  """Returns an array of all single target behavioural deviations.

  The target behavioural deviations are with respect to an information set.

  Args:
    num_actions: the integer of all actions that can be taken at that
      information set
    history: an array containing the prior actions played by the `player` to
      reach the information set.
    prior_legal_actions: a 2d array containing the legal actions for each
      preceeding state.

  Returns:
    an array of LocalDeviationWithTimeSelection objects that represent
    all (single target) behaviourial deviations that are realizable at the
    information set.
  """
  deviations = []
  if not history:
    internal = return_all_non_identity_internal_deviations(
        num_actions, [None], history
    )
    for i in internal:
      deviations.append(i)
  else:
    for deviation_info in range(len(history)):
      prior_possible_memory_actions = generate_all_action_permutations(
          [], prior_legal_actions[: deviation_info + 1]
      )
      memory_weights = np.concatenate(
          (np.ones(deviation_info), np.zeros(len(history) - deviation_info))
      )
      for prior_memory_actions in prior_possible_memory_actions:
        prior_memory_actions = np.concatenate((
            prior_memory_actions,
            np.zeros(len(history) - len(prior_memory_actions)),
        ))
        for _ in range(len(history) - len(prior_memory_actions)):
          prior_memory_actions.append(0)
        prior_memory_actions_cp = prior_memory_actions.copy()
        internal = return_all_non_identity_internal_deviations(
            num_actions, [memory_weights], prior_memory_actions_cp
        )
        for i in internal:
          deviations.append(i)

  return deviations


class LocalDeviationWithTimeSelection:
  """Comprised of a swap transformation.

  Comprised of a swap transformation that will be applied at the
  current information state, a memory weighting which describes
  the actions that are remembered and the memory action history
  (prior_memory_actions) that is remembered.
  Note that the "memory action history" might not equal the history in
  the case of some deviation types (e.g tips deviations).
  """

  # The swap transformation that will be compared to the unmodified strategy.
  # The transformation is applied at the memory state.
  local_swap_transform = attr.ib()

  # Which actions have been forgotten (0) or remembered (1) according
  # to the memory state.
  prior_actions_weight = attr.ib()

  # Which actions have been take according to the memory state
  prior_memory_actions = attr.ib()

  use_unmodified_history = attr.ib()

  def __init__(
      self,
      target,
      source,
      num_actions,
      prior_actions_weight,
      prior_memory_actions,
      is_external,
      use_unmodified_history=True,
  ):
    """Represents a swap transformation (either external and internal).

    Represents a swap transformation (either external and internal) for a given
    memory state.

    Args:
      target: the action that will be played when the deviation is triggered.
      source: the action that will trigger the target action when suggested
        (used only by internal deviations, i.e is_external = False).
      num_actions: the number of actions that can be played for this
        information state.
      prior_actions_weight: an array (the length of the game history)
        of the information state actions have been forgotten (0)
        or remembered (1) wrt to the memory state.
        This is represented numerically for possible experimentation with
        "partially forgotten" actions (i.e in the range (0,1)).
      prior_memory_actions: the preceeding actions upto the the information
        state (which the LocalDeviationWithTimeSelection is defined with respect
        to).
      is_external: a boolean use to determine whether this is an
        internal or external deviation.
      use_unmodified_history: a boolean used to indicate whether the provided
        memory_actions are the same as the information state it was derived
        from.
    """
    self.local_swap_transform = LocalSwapTransform(
        target, source, num_actions, is_external=is_external
    )
    self.prior_actions_weight = prior_actions_weight
    self.prior_memory_actions = prior_memory_actions
    self.use_unmodified_history = use_unmodified_history

  # If a pure strategy, a pure strategy will be returned (aka function works
  # for both actions and strategies as input).
  def deviate(self, strategy):
    """Returns a strategy array.
    
    Returns the strategy array given by deviating according to the
    'self.local_swap_transform.matrix_transform' matrix.

    Args:
      strategy: the strategy array to deviate from.

    Returns:
      the matrix product of the the matrix_transform and the provided strategy.
    """
    return self.local_swap_transform.deviate(strategy)

  def return_transform_matrix(self):
    """Returns a matrix_transform.

    Returns the matrix_transform of the associated `LocalSwapTransform` object.
    """
    return self.local_swap_transform.matrix_transform

  def player_deviation_reach_probability(
      self, prior_possible_action_probabilities
  ):
    """Calculate the probability of reaching the current memory state.

    Calculate the probability of reaching the current memory state
    provided the player played from the start of the game to this state.
    This is assuming that they play with their current strategy with the
    deviation applied.
    Args:
       prior_possible_action_probabilities: a 2d array of length [player's
         history]x[number of actions at that state]. These are the current
         strategies of the player, from start to end of their history.

    Returns:
       The reach probability of the current memory state.
    """
    if (
        self.prior_actions_weight is None
        or self.prior_memory_actions is None
        or prior_possible_action_probabilities is None
    ):
      return 1.0

    memory_action_probabilities = np.ones(len(self.prior_actions_weight))
    # Reconstruct memory probabilities from history provided to the deviation
    # to reach info set and the current memory probs.
    memory_weightings = self.prior_actions_weight.copy()
    if self.use_unmodified_history:
      for state in range(len(self.prior_memory_actions)):
        if self.prior_actions_weight[state] != 0:
          memory_action_probabilities[state] = (
              prior_possible_action_probabilities[state][
                  self.prior_memory_actions[state]
              ]
          )
        else:
          memory_action_probabilities[state] = 1
          memory_weightings[state] = 1

    path_probability = np.multiply(
        memory_weightings, memory_action_probabilities
    )
    memory_reach_probability = np.prod(path_probability)
    return memory_reach_probability

  def __eq__(self, other):
    return self.local_swap_transform == other.local_swap_transform

  def __hash__(self):
    return hash(self.local_swap_transform)


def return_all_non_identity_internal_deviations(
    num_actions, possible_prior_weights, prior_memory_actions
):
  """Returns all non-identity internal deviations."""
  deviations = []
  for prior_actions_weight in possible_prior_weights:
    for target in range(num_actions):
      for source in range(num_actions):
        if source != target:
          deviations.append(
              LocalDeviationWithTimeSelection(
                  target,
                  source,
                  num_actions,
                  prior_actions_weight,
                  prior_memory_actions,
                  False,
              )
          )
  return deviations


def return_all_internal_modified_deviations(
    num_actions,
    possible_prior_weights,
    possible_prior_memory_actions,
    prior_memory_actions,
):
  """Returns all internal deviations with modified memory actions."""
  deviations = []
  for prior_actions_weight in possible_prior_weights:
    try:
      modification_index = np.where(prior_actions_weight == 0)[0][0]
    except IndexError:
      modification_index = 0
    if modification_index == len(prior_memory_actions):
      for target in range(num_actions):
        for source in range(num_actions):
          if source != target:
            deviations.append(
                LocalDeviationWithTimeSelection(
                    target,
                    source,
                    num_actions,
                    prior_actions_weight,
                    prior_memory_actions,
                    False,
                )
            )
    else:
      previous_action = prior_memory_actions[modification_index]
      for alt_action in possible_prior_memory_actions[modification_index]:
        prior_memory_actions[modification_index] = alt_action
        for target in range(num_actions):
          for source in range(num_actions):
            if source != target:
              deviations.append(
                  LocalDeviationWithTimeSelection(
                      target,
                      source,
                      num_actions,
                      prior_actions_weight,
                      prior_memory_actions.copy(),
                      False,
                  )
              )
        prior_memory_actions[modification_index] = previous_action
  return deviations


def return_all_external_deviations(
    num_actions, possible_prior_weights, prior_memory_actions
):
  """Returns all external deviations."""
  deviations = []
  for prior_actions_weight in possible_prior_weights:
    for target in range(num_actions):
      deviations.append(
          LocalDeviationWithTimeSelection(
              target,
              target,
              num_actions,
              prior_actions_weight,
              prior_memory_actions,
              True,
          )
      )
  return deviations


# Modify last action as required
def return_all_external_modified_deviations(
    num_actions,
    possible_prior_weights,
    possible_prior_memory_actions,
    prior_memory_actions,
):
  """Returns all external deviations with modified memory actions."""
  deviations = []
  for prior_actions_weight in possible_prior_weights:
    try:
      modification_index = np.where(prior_actions_weight == 0)[0][0]
    except IndexError:
      modification_index = 0
    if modification_index == len(prior_memory_actions):
      for target in range(num_actions):
        deviations.append(
            LocalDeviationWithTimeSelection(
                target,
                target,
                num_actions,
                prior_actions_weight,
                prior_memory_actions,
                True,
            )
        )
    else:
      previous_action = prior_memory_actions[modification_index]
      for alt_action in possible_prior_memory_actions[modification_index]:
        prior_memory_actions[modification_index] = alt_action
        for target in range(num_actions):
          deviations.append(
              LocalDeviationWithTimeSelection(
                  target,
                  target,
                  num_actions,
                  prior_actions_weight,
                  prior_memory_actions.copy(),
                  True,
              )
          )
        prior_memory_actions[modification_index] = previous_action
  return deviations


def return_identity_deviation(
    num_actions, possible_prior_weights, prior_memory_actions
):
  deviations = []
  for prior_actions_weight in possible_prior_weights:
    deviations.append(
        LocalDeviationWithTimeSelection(
            0, 0, num_actions, prior_actions_weight, prior_memory_actions, False
        )
    )
  return deviations


# A swap transformation given by the matrix_transform for an information state.
# Of actions_num size.
class LocalSwapTransform:
  """Represents a swap transformation (both external and internal).

  Represents a swap transformation (both external and internal)
  for an information state for a certain number of actions.
  """

  source_action = attr.ib()
  target_action = attr.ib()
  matrix_transform = attr.ib()
  actions_num = attr.ib()
  is_external = attr.ib()

  def __init__(self, target, source, actions_num, is_external=True):
    """Creates a matrix transformation.

    Creates the matrix transformation describing the swap transformation
    and initalises variables.

    Args:
      target: the action that will be played when the deviation is triggered.
      source: the action that triggers a swap to the target action
        (used only by internal deviations, i.e is_external = False)
      actions_num: the number of actions that can be played for this
        information state.
      is_external: determine whether to create an internal or external
        deviation.
    """
    self.source_action = source
    self.target_action = target
    self.actions_num = actions_num
    if is_external:
      self.source_action = None
      self.matrix_transform = np.zeros((actions_num, actions_num))
      self.matrix_transform[target] = np.ones(actions_num)
    else:
      self.matrix_transform = np.eye(actions_num)
      self.matrix_transform[target][source] = 1
      self.matrix_transform[source][source] = 0

  def __repr__(self) -> str:
    return (
        "Swapping from Action: "
        + str(self.source_action)
        + " to Action: "
        + str(self.target_action)
    )

  def __eq__(self, other) -> bool:
    return (
        self.source_action == other.source_action
        and self.target_action == other.target_action
        and self.actions_num == other.actions_num
    )

  def __hash__(self):
    return hash(
        f"{str(self.source_action)} {str(self.target_action)}                "
        f" {str(self.actions_num)} {str(self.is_external)}"
    )

  def deviate(self, strategy):
    """Returns a strategy array.

    Returns the strategy array given by deviating according to
    'self.matrix_transform' matrix.

    Args:
      strategy: the strategy array to deviate from.

    Returns:
      the matrix product of the the matrix_transform and the provided strategy.
    """
    return np.matmul(self.matrix_transform, strategy)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for go."""

import json

from absl.testing import absltest
import pyspiel

go = pyspiel.go


class GamesGoTest(absltest.TestCase):

  def test_json(self):
    game = pyspiel.load_game("go")
    state = game.new_initial_state()
    state_struct = state.to_struct()
    self.assertEqual(state_struct.current_player, "B")
    self.assertEqual(state_struct.board_grid[0][0]["a1"], "EMPTY")
    state.apply_action(0)
    state_json = state.to_json()
    state_dict = json.loads(state_json)
    state_struct = go.GoStateStruct(state_json)
    self.assertEqual(state_struct.current_player, "W")
    self.assertEqual(state_dict["current_player"], "W")
    self.assertEqual(state_struct.board_grid[0][0]["a1"], "B")
    self.assertEqual(state_struct.to_json(),
                     json.dumps(state_dict, separators=(",", ":")))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Find Nash equilibria for constant- or general-sum 2-player games.

Non-matrix games are handled by computing the normal (bimatrix) form.
The algorithms used are:
* direct computation of pure equilibria.
* linear programming to find equilibria for constant-sum games.
* iterated dominance to reduce the action space.
* reverse search vertex enumeration (if using lrsnash) to find all general-sum
  equilibria.
* support enumeration (if using nashpy) to find all general-sum equilibria.
* Lemke-Howson enumeration (if using nashpy) to find one general-sum
  equilibrium.
The general-sum mixed-equilibrium algorithms are likely to work well for tens of
actions, but less likely to scale beyond that.
"""

import fractions
import os
import subprocess
import tempfile
import warnings

import nashpy
import numpy as np


@np.vectorize
def to_fraction_str(x, lrsnash_max_denom):
  return str(fractions.Fraction(x).limit_denominator(lrsnash_max_denom))


def lrs_solve(row_payoffs, col_payoffs, lrsnash_max_denom, lrsnash_path):
  """Find all Nash equilibria using the lrsnash solver.

  `lrsnash` uses reverse search vertex enumeration on rational polytopes.
  For more info, see: http://cgm.cs.mcgill.ca/~avis/C/lrslib/USERGUIDE.html#nash

  Args:
    row_payoffs: payoffs for row player
    col_payoffs: payoffs for column player
    lrsnash_max_denom: maximum denominator
    lrsnash_path: path for temporary files

  Yields:
    (row_mixture, col_mixture), numpy vectors of float64s.
  """
  num_rows, num_cols = row_payoffs.shape
  game_file, game_file_path = tempfile.mkstemp()
  try:
    game_file = os.fdopen(game_file, "w")

    # write dimensions
    game_file.write("%d %d\n\n" % (num_rows, num_cols))

    # write row-player payoff matrix as fractions
    for row in range(num_rows):
      game_file.write(
          " ".join(to_fraction_str(row_payoffs[row], lrsnash_max_denom)) + "\n")
    game_file.write("\n")

    # write col-player payoff matrix as fractions
    for row in range(num_rows):
      game_file.write(
          " ".join(to_fraction_str(col_payoffs[row], lrsnash_max_denom)) + "\n")
    game_file.write("\n")
    game_file.close()
    lrs = subprocess.Popen([lrsnash_path or "lrsnash", "-s", game_file_path],
                           stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
    col_mixtures = []
    for line in lrs.stdout:
      if len(line) <= 1 or line[:1] == b"*":
        continue
      line = np.asarray(
          [fractions.Fraction(x) for x in line.decode().split()],
          dtype=np.float64,
      )
      if line[0] == 2:  # col-player
        col_mixtures.append(line[1:-1])
      else:  # row-player
        row_mixture = line[1:-1]
        # row-mixture forms a Nash with every col-mixture listed directly above
        for col_mixture in col_mixtures:
          yield (row_mixture, col_mixture)
        col_mixtures = []
  finally:
    os.remove(game_file_path)


def lemke_howson_solve(row_payoffs, col_payoffs):
  """Find Nash equilibria using the Lemke-Howson algorithm.

  The algorithm is not guaranteed to find all equilibria. Also it can yield
  wrong answers if the game is degenerate (but raises warnings in that case).
  Args:
    row_payoffs: payoffs for row player
    col_payoffs: payoffs for column player
  Yields:
    (row_mixture, col_mixture), numpy vectors of float64s.
  """

  showwarning = warnings.showwarning
  warned_degenerate = [False]

  def showwarning_check_degenerate(message, *args, **kwargs):
    if "Your game could be degenerate." in str(message):
      warned_degenerate[0] = True
    showwarning(message, *args, **kwargs)

  try:
    warnings.showwarning = showwarning_check_degenerate
    for row_mixture, col_mixture in nashpy.Game(
        row_payoffs, col_payoffs).lemke_howson_enumeration():
      if warned_degenerate[0]:
        # attempt to discard obviously-wrong results
        if (row_mixture.shape != row_payoffs.shape[:1] or
            col_mixture.shape != row_payoffs.shape[1:]):
          warnings.warn("Discarding ill-shaped solution.")
          continue
        if (not np.isfinite(row_mixture).all() or
            not np.isfinite(col_mixture).all()):
          warnings.warn("Discarding non-finite solution.")
          continue
      yield row_mixture, col_mixture
  finally:
    warnings.showwarning = showwarning



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods to compute the core based on stochastic saddle point programming.

This file contains methods to compute the core using a Lagrangian formulation
referred to in Gemp et al AAMAS '24:
"Approximating the Core via Iterative Coalition Sampling"

TODO:
- add a link to arXiv when it's live 
- add the least core algorithm to the algorithms.md also when that link is live
"""

import dataclasses
import functools
import itertools
import time
from typing import Any, Dict, Tuple

from absl import logging
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python.coalitional_games import coalitional_game


@dataclasses.dataclass(frozen=True)
class LeastCoreValue:
  payoff: np.ndarray
  lcv: float
  duration: float
  meta: Dict[Any, Any]


def compute_least_core_value(
    cvc: coalitional_game.CoalitionalGame, alg_config
) -> LeastCoreValue:
  """Computes the least core value of a game."""
  opt_primal = optax.adam(learning_rate=alg_config.init.lr_primal)
  opt_dual = optax.adam(learning_rate=alg_config.init.lr_dual)
  evaluation_iterations = alg_config.eval.evaluation_iterations
  evaluate_every = 2 * alg_config.solve.n_iter  # do not evaluate
  cl = CoreLagrangian(cvc, opt_primal, opt_dual)
  payoffs, epsilons, _, duration = cl.solve(
      evaluation_iterations=evaluation_iterations,
      evaluate_every=evaluate_every,
      **alg_config.solve,
  )
  lcvs = np.full(payoffs.shape[0], np.inf)
  payoff = payoffs[-1]
  lcv = np.inf
  for i in range(payoffs.shape[0]):
    payoff = payoffs[i]
    epsilon = epsilons[i]
    max_violation = payoff_evaluation(
        cvc, payoff, epsilon, evaluation_iterations)
    lcv = epsilon + max_violation
    lcvs[i] = lcv
  meta = dict(payoffs=payoffs, epsilons=epsilons, lcvs=lcvs)
  return LeastCoreValue(payoff, lcv, duration, meta)


def payoff_evaluation(
    cv_calc: coalitional_game.CoalitionalGame,
    payoffs: np.ndarray,
    epsilon: float,
    batch_size: int,
    max_exponent: int = 13,
) -> float:
  """Evaluate deficit over a set of random coalitions.

  Args:
    cv_calc: the game to work on
    payoffs: the payoff vector to test
    epsilon: desired approximation of the epsilon-core
    batch_size: number of random coalitions to sample
    max_exponent: examine at maxixum 2**max_exponent constraints in one batch
      default 13, assume 2**13 ~ 10k coalitions is mem limit for single batch

  Returns:
    Expected loss, relu(deficit), over random batch of coalitions
  """
  max_batch = 2**max_exponent
  num_players = cv_calc.num_players()
  violation = 0.
  if batch_size >= 2**num_players:
    num_suffix_repeats = min(max_exponent, num_players)
    num_prefix_repeats = max(0, num_players - num_suffix_repeats)
    zo = [0, 1]
    suffix = np.array(list(itertools.product(zo, repeat=num_suffix_repeats)))
    prefixes = itertools.product(zo, repeat=num_prefix_repeats)
    for prefix in prefixes:
      if prefix:
        prefix_rep = np.repeat([prefix], suffix.shape[0], axis=0)
        coalitions = np.concatenate([prefix_rep, suffix], axis=1)
      else:
        coalitions = suffix
      batch_contributions = cv_calc.coalition_values(coalitions)
      batch_payouts = np.dot(coalitions, payoffs)
      batch_deficits = batch_contributions - batch_payouts - epsilon
      batch_deficits = np.clip(batch_deficits, 0., np.inf)
      violation = max(violation, np.max(batch_deficits))
  else:
    q, r = divmod(batch_size, max_batch)
    num_loops = q + (r > 0)
    for _ in range(num_loops):
      coalitions = np.random.randint(2, size=(max_batch, num_players))
      batch_contributions = cv_calc.coalition_values(coalitions)
      batch_payouts = np.dot(coalitions, payoffs)
      batch_deficits = batch_contributions - batch_payouts - epsilon
      batch_deficits = np.clip(batch_deficits, 0., np.inf)
      violation = max(violation, np.max(batch_deficits))
  return float(violation)


class CoreSolver():
  """Find an epsilon-core."""

  def __init__(self,
               cvc: coalitional_game.CoalitionalGame):
    self.cvc = cvc
    self.num_players = cvc.num_players()
    # we assume grand_coalition is optimal coalition
    grand_coalition = np.full(cvc.num_players(), 1, dtype=np.int32)
    self.grand_coalition_value = cvc.coalition_value(grand_coalition)

    self.payoffs = None
    self.losses = None
    self.max_deficits = None
    self.evaluation_losses = None

  def logits_to_payoff(self, logits):
    logits_ext = jnp.append(logits, 0.)
    payoff = jax.nn.softmax(logits_ext)
    payoff *= self.grand_coalition_value
    return payoff

  def loss_deficit(self, current_payoff, coalitions, coalition_values, epsilon):
    """Compute Mean Loss and Max Deficit."""
    coalition_payment = jnp.dot(coalitions, current_payoff)
    deficit = coalition_values - epsilon - coalition_payment
    coalition_size = jnp.sum(coalitions, axis=1)
    weight = 1.0 / jnp.clip(coalition_size, 1, self.num_players)
    losses = 0.5 * weight * jax.nn.relu(deficit) ** 2.0
    return jnp.mean(losses, axis=0), jnp.max(jax.nn.relu(deficit))


class CoreOptimization(CoreSolver):
  """Find an epsilon-core via optimization."""

  def __init__(self,
               cvc: coalitional_game.CoalitionalGame,
               opt,
               epsilon):
    super().__init__(cvc)
    self.opt = opt
    self.epsilon = epsilon

  @functools.partial(jax.jit, static_argnums=[0])
  def loss(self, params, data):
    """Compute Loss."""
    current_payoff = params
    coalitions, coalition_values = data

    return self.loss_deficit(current_payoff, coalitions, coalition_values,
                             self.epsilon)

  @functools.partial(jax.jit, static_argnums=[0])
  def update_step(self, params, data, opt_state):
    """GD update step."""

    # data = (coalitions, coalition_values)

    # Convert losses into pure functions.
    loss_fn = lambda p: self.loss(p, data)[0]

    # Compute saddle-point gradients.
    grads_fn = jax.grad(loss_fn, argnums=0)
    grads = grads_fn(params)

    updates, opt_state = self.opt.update(grads, opt_state, params)

    params = optax.apply_updates(params, updates)

    params = jnp.clip(params, 0, np.inf)
    scale = self.grand_coalition_value / jnp.sum(params)
    params = params * scale

    return params, opt_state

  def solve(self, n_iter: int, batch_size: int = 100,
            save_every: int = 2,
            evaluate_every: int = 2, evaluation_iterations: int = 100,
            seed: int = 0
            ) -> Tuple[np.ndarray, np.ndarray, float]:
    """Find a least-core via Lagrange multipliers.

    Additional optimization metrics are stored as class variables:
      self.payoffs
      self.losses
      self.max_deficits
      self.evaluation_losses
      self.duration

    Args:
      n_iter: number of iterations
      batch_size: number of constraints to evaluate at each step
      save_every: int, how often to record optimization variables
      evaluate_every: int, how often to evaluate the max constraint violation
      evaluation_iterations: int, how many constraints to measure violations
        for, if number if less than number of coalitions a batch of constraints
        is sampled randomly. otherwise, all constraints are evaluated
      seed: int, for sampling minibatches of constraints

    Returns:
      payoffs over training
      max deficit over training
      runtime duration (sec)
    """

    qe, re = divmod(n_iter, evaluate_every)
    num_eval = qe + (re > 0)
    qs, rs = divmod(n_iter, save_every)
    num_save = qs + (rs > 0)

    max_violations = np.empty(num_eval, dtype=np.float32)
    losses = np.empty(num_save, dtype=np.float32)
    max_deficits = np.empty(num_save, dtype=np.float32)
    payoffs = np.empty((num_save, self.num_players), dtype=np.float32)

    scale = self.grand_coalition_value / self.num_players
    grand_coalition = np.full(self.num_players, 1, dtype=np.int32)
    current_payoff = jnp.array(grand_coalition * scale)
    params = current_payoff

    opt_state = self.opt.init(params)

    logging.debug('Uniform payoff %s', current_payoff)

    rng = jax.random.PRNGKey(seed)

    start = time.time()
    for iter_id in range(n_iter):
      if batch_size < 2**self.num_players:
        rng, key = jax.random.split(rng, 2)
        coalitions = jax.random.randint(key,
                                        shape=(batch_size, self.num_players),
                                        minval=0,
                                        maxval=2,
                                        dtype=jnp.int32)
      else:
        prod_space = itertools.product([0, 1], repeat=self.num_players)
        coalitions = np.stack(list(prod_space))
      coalition_values = self.cvc.coalition_values(np.array(coalitions))

      data = (coalitions, coalition_values)
      loss, max_deficit = self.loss(params, data)
      params, opt_state = self.update_step(params, data, opt_state)

      # Done updating, save if needed
      if iter_id % save_every == 0:
        logging.debug('Saving...')
        idx = iter_id // save_every
        losses[idx] = loss
        max_deficits[idx] = max_deficit
        current_payoff = params
        payoffs[idx] = current_payoff
        logging.debug('Loss was %f, Max deficit was %f, New payoff %s',
                      loss, max_deficit, current_payoff)

      # Done updating, evaluate if needed
      if (evaluate_every < n_iter) and (iter_id % evaluate_every == 0):
        logging.debug('Evaluating...')
        estimated_loss = payoff_evaluation(
            self.cvc,
            current_payoff,
            self.epsilon,
            evaluation_iterations,
        )
        max_violations[iter_id // evaluate_every] = estimated_loss
        logging.debug('Estimated loss %f', estimated_loss)
    end = time.time()
    duration = end - start

    self.payoffs = np.array(payoffs)
    self.losses = np.array(losses)
    self.max_deficits = np.array(max_deficits)
    self.max_violations = np.array(max_violations)
    self.duration = duration

    return (np.array(payoffs),
            np.array(max_deficits),
            duration)


class CoreOptimizationLogits(CoreSolver):
  """Find an epsilon-core via optimization over logits."""

  def __init__(self,
               cvc: coalitional_game.CoalitionalGame,
               opt,
               epsilon):
    super().__init__(cvc)
    self.opt = opt
    self.epsilon = epsilon

  @functools.partial(jax.jit, static_argnums=[0])
  def loss(self, params, data):
    """Compute Loss."""
    current_payoff = self.logits_to_payoff(params)
    coalitions, coalition_values = data

    return self.loss_deficit(current_payoff, coalitions, coalition_values,
                             self.epsilon)

  @functools.partial(jax.jit, static_argnums=[0])
  def update_step(self, params, data, opt_state):
    """GD update step."""

    # data = (coalitions, coalition_values)

    # Convert losses into pure functions.
    loss_fn = lambda p: self.loss(p, data)[0]

    # Compute saddle-point gradients.
    grads_fn = jax.grad(loss_fn, argnums=0)
    grads = grads_fn(params)

    updates, opt_state = self.opt.update(grads, opt_state, params)

    params = optax.apply_updates(params, updates)

    return params, opt_state

  def solve(self, n_iter: int, batch_size: int = 100,
            save_every: int = 2,
            evaluate_every: int = 2, evaluation_iterations: int = 100,
            seed: int = 0
            ) -> Tuple[np.ndarray, np.ndarray, float]:
    """Find a least-core via Lagrange multipliers.

    Additional optimization metrics are stored as class variables:
      self.payoffs
      self.losses
      self.max_deficits
      self.evaluation_losses
      self.duration

    Args:
      n_iter: number of iterations
      batch_size: number of constraints to evaluate at each step
      save_every: int, how often to record optimization variables
      evaluate_every: int, how often to evaluate the max constraint violation
      evaluation_iterations: int, how many constraints to measure violations
        for, if number if less than number of coalitions a batch of constraints
        is sampled randomly. otherwise, all constraints are evaluated
      seed: int, for sampling minibatches of constraints

    Returns:
      payoffs over training
      max deficit over training
      runtime duration (sec)
    """

    qe, re = divmod(n_iter, evaluate_every)
    num_eval = qe + (re > 0)
    qs, rs = divmod(n_iter, save_every)
    num_save = qs + (rs > 0)

    max_violations = np.empty(num_eval, dtype=np.float32)
    losses = np.empty(num_save, dtype=np.float32)
    max_deficits = np.empty(num_save, dtype=np.float32)
    payoffs = np.empty((num_save, self.num_players), dtype=np.float32)

    current_logits = jnp.zeros(self.num_players - 1, dtype=jnp.float32)
    current_payoff = np.asarray(self.logits_to_payoff(current_logits))
    params = current_logits

    opt_state = self.opt.init(params)

    logging.debug('Uniform payoff %s', current_payoff)

    rng = jax.random.PRNGKey(seed)

    start = time.time()
    for iter_id in range(n_iter):
      if batch_size < 2**self.num_players:
        rng, key = jax.random.split(rng, 2)
        coalitions = jax.random.randint(key,
                                        shape=(batch_size, self.num_players),
                                        minval=0,
                                        maxval=2,
                                        dtype=jnp.int32)
      else:
        prod_space = itertools.product([0, 1], repeat=self.num_players)
        coalitions = np.stack(list(prod_space))
      coalition_values = self.cvc.coalition_values(np.array(coalitions))

      data = (coalitions, coalition_values)
      loss, max_deficit = self.loss(params, data)
      params, opt_state = self.update_step(params, data, opt_state)

      # Done updating, save if needed
      if iter_id % save_every == 0:
        logging.debug('Saving...')
        idx = iter_id // save_every
        losses[idx] = loss
        max_deficits[idx] = max_deficit
        current_logits = params
        current_payoff = np.asarray(self.logits_to_payoff(current_logits))
        payoffs[idx] = current_payoff
        logging.debug('Loss was %f, Max deficit was %f, New payoff %s',
                      loss, max_deficit, current_payoff)

      # Done updating, evaluate if needed
      if (evaluate_every < n_iter) and (iter_id % evaluate_every == 0):
        logging.debug('Evaluating...')
        estimated_loss = payoff_evaluation(
            self.cvc,
            current_payoff,
            self.epsilon,
            evaluation_iterations,
        )
        max_violations[iter_id // evaluate_every] = estimated_loss
        logging.debug('Estimated loss %f', estimated_loss)
    end = time.time()
    duration = end - start
    self.payoffs = np.array(payoffs)
    self.losses = np.array(losses)
    self.max_deficits = np.array(max_deficits)
    self.max_violations = np.array(max_violations)
    self.duration = duration

    return (np.array(payoffs),
            np.array(max_deficits),
            duration)


class CoreLagrangian(CoreSolver):
  """Find a least-core via Lagrange multipliers."""

  def __init__(self,
               cvc: coalitional_game.CoalitionalGame,
               opt_primal,
               opt_dual):
    super().__init__(cvc)
    self.opt_primal = opt_primal
    self.opt_dual = opt_dual

    current_logits_keys = ['current_logits' for _ in range(self.num_players)]
    keys_primal = {'current_logits': current_logits_keys,
                   'epsilon': 'epsilon'}
    keys_dual = {'mu': 'mu'}
    self.keys = (keys_primal, keys_dual)
    self.nonnegative_keys = ('epsilon', 'mu')

    self.epsilons = None
    self.mus = None
    self.lagrangians = None

  @functools.partial(jax.jit, static_argnums=[0])
  def lagrangian(self, primal, dual, data):
    """Compute Lagrangian."""
    current_logits, epsilon = primal['current_logits'], primal['epsilon']
    mu = dual['mu']
    coalitions, coalition_values, gamma_adj = data

    current_payoff = self.logits_to_payoff(current_logits)
    mean_loss, max_deficit = self.loss_deficit(current_payoff,
                                               coalitions,
                                               coalition_values,
                                               epsilon)
    lagrangian = epsilon + mu * (mean_loss - gamma_adj)
    lagrangian = jnp.sum(lagrangian)  # just for converting (1,) array to scalar
    return lagrangian, (mean_loss, max_deficit)

  @functools.partial(jax.jit, static_argnums=[0])
  def update_step(self, params, data, opt_state):
    """SimGD update step."""

    # data = (coalitions, coalition_values, gamma_adj)
    params_primal, params_dual = params
    opt_state_primal, opt_state_dual = opt_state

    # Convert losses into pure functions.
    loss_primal_fn = lambda p, d: self.lagrangian(p, d, data)[0]
    loss_dual_fn = lambda p, d: -self.lagrangian(p, d, data)[0]

    # Compute saddle-point gradients.
    grads_primal_fn = jax.grad(loss_primal_fn, argnums=0)
    grads_primal = grads_primal_fn(params_primal, params_dual)
    grads_dual_fn = jax.grad(loss_dual_fn, argnums=1)
    grads_dual = grads_dual_fn(params_primal, params_dual)

    updates_primal, opt_state_primal = self.opt_primal.update(grads_primal,
                                                              opt_state_primal,
                                                              params_primal)
    updates_dual, opt_state_dual = self.opt_dual.update(grads_dual,
                                                        opt_state_dual,
                                                        params_dual)

    params_primal = optax.apply_updates(params_primal, updates_primal)
    params_dual = optax.apply_updates(params_dual, updates_dual)

    params = (params_primal, params_dual)
    opt_state = (opt_state_primal, opt_state_dual)

    clip = (
        lambda x, k: jnp.clip(x, 0, np.inf) if k in self.nonnegative_keys else x
    )
    params = jax.tree_util.tree_map(clip, params, self.keys)

    return params, opt_state

  def solve(self, n_iter: int, batch_size: int = 100, gamma: float = 1e-2,
            mu_init: float = 1000.,
            save_every: int = 2,
            evaluate_every: int = 2, evaluation_iterations: int = 100,
            seed: int = 0,
            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float]:
    """Find a least-core via Lagrange multipliers.

    Additional optimization metrics are stored as class variables:
      self.payoffs
      self.epsilons
      self.mus
      self.lagrangians
      self.losses
      self.max_deficits
      self.evaluation_losses
      self.duration

    Args:
      n_iter: number of iterations
      batch_size: number of constraints to evaluate at each step
      gamma: float, slack allowed in core constraints
      mu_init: float, initialize the lagrange multiplier to this value
      save_every: int, how often to record optimization variables
      evaluate_every: int, how often to evaluate the max constraint violation
      evaluation_iterations: int, how many constraints to measure violations
        for, if number if less than number of coalitions a batch of constraints
        is sampled randomly. otherwise, all constraints are evaluated
      seed: int, for sampling minibatches of constraints

    Returns:
      payoffs over training
      epsilon over training
      max deficit over training
      runtime duration (sec)
    """

    qe, re = divmod(n_iter, evaluate_every)
    num_eval = qe + (re > 0)
    qs, rs = divmod(n_iter, save_every)
    num_save = qs + (rs > 0)

    max_violations = np.empty(num_eval, dtype=np.float32)
    lagrangians = np.empty(num_save, dtype=np.float32)
    losses = np.empty(num_save, dtype=np.float32)
    max_deficits = np.empty(num_save, dtype=np.float32)
    epsilons = np.empty(num_save, dtype=np.float32)
    payoffs = np.empty((num_save, self.num_players), dtype=np.float32)
    mus = np.empty(num_save, dtype=np.float32)

    current_logits = jnp.zeros(self.num_players - 1, dtype=jnp.float32)
    epsilon = self.grand_coalition_value * jnp.ones(1, dtype=jnp.float32)
    mu = jnp.ones(1, dtype=jnp.float32) * mu_init

    params_primal = {'current_logits': current_logits,
                     'epsilon': epsilon}
    params_dual = {'mu': mu}
    params = (params_primal, params_dual)

    opt_state_primal = self.opt_primal.init(params_primal)
    opt_state_dual = self.opt_dual.init(params_dual)
    opt_state = (opt_state_primal, opt_state_dual)

    current_payoff = np.asarray(self.logits_to_payoff(current_logits))
    logging.debug('Uniform payoff %s', current_payoff)

    if self.num_players < 30:
      gamma_adj = gamma**2.0 / (2**self.num_players - 1)
    else:
      # Set arbitrary value if the above would result in a too tiny number.
      gamma_adj = 1e-6

    rng = jax.random.PRNGKey(seed)

    start = time.time()
    for iter_id in range(n_iter):
      if batch_size < 2**self.num_players:
        rng, key = jax.random.split(rng, 2)
        coalitions = jax.random.randint(key,
                                        shape=(batch_size, self.num_players),
                                        minval=0,
                                        maxval=2,
                                        dtype=jnp.int32)
      else:
        prod_space = itertools.product([0, 1], repeat=self.num_players)
        coalitions = np.stack(list(prod_space))
      coalition_values = self.cvc.coalition_values(np.array(coalitions))

      data = (coalitions, coalition_values, gamma_adj)
      lagrangian, (loss, max_deficit) = self.lagrangian(*params, data)
      params, opt_state = self.update_step(params, data, opt_state)

      params_primal, params_dual = params

      # Done updating, save if needed
      if iter_id % save_every == 0:
        logging.debug('Saving...')
        idx = iter_id // save_every
        lagrangians[idx] = lagrangian
        losses[idx] = loss
        max_deficits[idx] = max_deficit
        epsilons[idx] = params_primal['epsilon'].item()
        mus[idx] = params_dual['mu'].item()
        current_payoff = np.asarray(self.logits_to_payoff(
            params_primal['current_logits']))
        payoffs[idx] = current_payoff
        logging.debug('Loss was %f, Max deficit was %f, New payoff %s',
                      loss, max_deficit, current_payoff)

      # Done updating, evaluate if needed
      if (evaluate_every < n_iter) and (iter_id % evaluate_every == 0):
        logging.debug('Evaluating...')
        estimated_loss = payoff_evaluation(
            self.cvc,
            current_payoff,
            params_primal['epsilon'].item(),
            evaluation_iterations,
        )
        max_violations[iter_id // evaluate_every] = estimated_loss
        logging.debug('Estimated loss %f', estimated_loss)
    end = time.time()
    duration = end - start

    self.payoffs = np.array(payoffs)
    self.epsilons = np.array(epsilons)
    self.mus = np.array(mus)
    self.lagrangians = np.array(lagrangians)
    self.losses = np.array(losses)
    self.max_deficits = np.array(max_deficits)
    self.max_violations = np.array(max_violations)
    self.duration = duration

    return (np.array(payoffs),
            np.array(epsilons),
            np.array(max_deficits),
            duration)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation of iterated prisoner's dilemma.

This is primarily here to demonstrate simultaneous-move games in Python.
"""

import enum

import numpy as np

import pyspiel

_NUM_PLAYERS = 2
_DEFAULT_PARAMS = {"termination_probability": 0.125, "max_game_length": 9999}
_PAYOFF = [[5, 0], [10, 1]]

_GAME_TYPE = pyspiel.GameType(
    short_name="python_iterated_prisoners_dilemma",
    long_name="Python Iterated Prisoner's Dilemma",
    dynamics=pyspiel.GameType.Dynamics.SIMULTANEOUS,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=False,
    provides_information_state_tensor=False,
    provides_observation_string=False,
    provides_observation_tensor=False,
    provides_factored_observation_string=False,
    parameter_specification=_DEFAULT_PARAMS)


class Action(enum.IntEnum):
  COOPERATE = 0
  DEFECT = 1


class Chance(enum.IntEnum):
  CONTINUE = 0
  STOP = 1


class IteratedPrisonersDilemmaGame(pyspiel.Game):
  """The game, from which states and observers can be made."""

  # pylint:disable=dangerous-default-value
  def __init__(self, params=_DEFAULT_PARAMS):
    max_game_length = params["max_game_length"]
    super().__init__(
        _GAME_TYPE,
        pyspiel.GameInfo(
            num_distinct_actions=2,
            max_chance_outcomes=2,
            num_players=2,
            min_utility=np.min(_PAYOFF) * max_game_length,
            max_utility=np.max(_PAYOFF) * max_game_length,
            utility_sum=None,
            max_game_length=max_game_length,
        ),
        params,
    )
    self._termination_probability = params["termination_probability"]

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return IteratedPrisonersDilemmaState(self, self._termination_probability)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    return IteratedPrisonersDilemmaObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
        params)


class IteratedPrisonersDilemmaState(pyspiel.State):
  """Current state of the game."""

  def __init__(self, game, termination_probability):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._current_iteration = 1
    self._termination_probability = termination_probability
    self._is_chance = False
    self._game_over = False
    self._rewards = np.zeros(_NUM_PLAYERS)
    self._returns = np.zeros(_NUM_PLAYERS)

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every simultaneous-move game with chance.

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self._game_over:
      return pyspiel.PlayerId.TERMINAL
    elif self._is_chance:
      return pyspiel.PlayerId.CHANCE
    else:
      return pyspiel.PlayerId.SIMULTANEOUS

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    assert player >= 0
    return [Action.COOPERATE, Action.DEFECT]

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self._is_chance
    return [(Chance.CONTINUE, 1 - self._termination_probability),
            (Chance.STOP, self._termination_probability)]

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    # This is not called at simultaneous-move states.
    assert self._is_chance and not self._game_over
    self._current_iteration += 1
    self._is_chance = False
    self._game_over = (action == Chance.STOP)
    if self._current_iteration > self.get_game().max_game_length():
      self._game_over = True

  def _apply_actions(self, actions):
    """Applies the specified actions (per player) to the state."""
    assert not self._is_chance and not self._game_over
    self._is_chance = True
    self._rewards[0] = _PAYOFF[actions[0]][actions[1]]
    self._rewards[1] = _PAYOFF[actions[1]][actions[0]]
    self._returns += self._rewards

  def _action_to_string(self, player, action):
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      return Chance(action).name
    else:
      return Action(action).name

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._game_over

  def rewards(self):
    """Reward at the previous step."""
    return self._rewards

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    return self._returns

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    return (f"p0:{self.action_history_string(0)} "
            f"p1:{self.action_history_string(1)}")

  def action_history_string(self, player):
    return "".join(
        self._action_to_string(pa.player, pa.action)[0]
        for pa in self.full_history()
        if pa.player == player)


class IteratedPrisonersDilemmaObserver:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, iig_obs_type, params):
    """Initializes an empty observation tensor."""
    assert not bool(params)
    self.iig_obs_type = iig_obs_type
    self.tensor = None
    self.dict = {}

  def set_from(self, state, player):
    pass

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    if self.iig_obs_type.public_info:
      return (f"us:{state.action_history_string(player)} "
              f"op:{state.action_history_string(1 - player)}")
    else:
      return None


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, IteratedPrisonersDilemmaGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean-Field PSRO examples."""

from absl import app
from absl import flags
from absl import logging

from open_spiel.python.mfg.algorithms import correlated_equilibrium
from open_spiel.python.mfg.algorithms import mf_psro
from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import hedge
from open_spiel.python.mfg.algorithms.regret import polynomial_weights
from open_spiel.python.mfg.algorithms.regret import regret_matching
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
from open_spiel.python.mfg.games import dynamic_routing  # pylint: disable=unused-import
from open_spiel.python.mfg.games import normal_form_game  # pylint: disable=unused-import
from open_spiel.python.mfg.games import predator_prey  # pylint: disable=unused-import
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("game_name", "python_mfg_predator_prey",
                    "Name of the game.")
flags.DEFINE_integer(
    "regret_steps_per_step",
    1000,
    "number of runs to average value function over.",
)
flags.DEFINE_integer(
    "value_estimation_n", 1, "number of runs to average value function over."
)
flags.DEFINE_string(
    "value_estimator", "sampled", "Best Response type : `ce` or `cce`."
)
flags.DEFINE_string(
    "regret_minimizer",
    "hedge",
    "Which regret minimization algorithm to use : `rm` for"
    "Regret Matching, `hedge` for Hedge, `poly` for Polynomial "
    "Weights.",
)
flags.DEFINE_integer("n_iter", 1000, "Num PSRO iterations.")
flags.DEFINE_integer("compress_every", 1, "Compress every")
flags.DEFINE_float("compress_lbd", 0.0, "Compression lambda.")
flags.DEFINE_float("eta", None, "Polynomial Weight algorithm eta.")
flags.DEFINE_string(
    "best_responder", "cce", "Best Response type : `ce` or `cce`."
)
flags.DEFINE_bool(
    "compute_internal_regret",
    False,
    "Compute internal (Or external if False) regret",
)
flags.DEFINE_bool("compute_ce_gap", False, "Compute `ce_gap`")
flags.DEFINE_integer("seed", 1, "Seed value.")

GAME_SETTINGS = {
    "mfg_crowd_modelling_2d": {
        "only_distribution_reward": False,
        "forbidden_states": "[0|0;0|1]",
        "initial_distribution": "[0|2;0|3]",
        "initial_distribution_value": "[0.5;0.5]",
    }
}


def main(unused_argv):
  logging.info("Loading %s", FLAGS.game_name)
  mfg_game = pyspiel.load_game(
      FLAGS.game_name, GAME_SETTINGS.get(FLAGS.game_name, {})
  )

  eta = FLAGS.eta
  regret_steps_per_step = FLAGS.regret_steps_per_step

  best_responder = FLAGS.best_responder
  compute_ce_gap = FLAGS.compute_ce_gap
  compute_internal_regret = FLAGS.compute_internal_regret

  if FLAGS.value_estimator == "sampled":
    value_estimator = utils.sample_value
  elif FLAGS.value_estimator == "exact":
    value_estimator = utils.get_exact_value
  else:
    raise NameError(
        "Unknown value estimator {}. Valid names are `sampled`, `exact`."
        .format(FLAGS.value_estimator)
    )

  if FLAGS.regret_minimizer == "hedge":
    regret_minimizer = hedge.Hedge(
        mfg_game,
        [],
        eta,
        regret_steps_per_step,
        compress_nus=True,
        compress_every=FLAGS.compress_every,
        compress_lbd=FLAGS.compress_lbd,
        value_estimator=value_estimator,
        value_estimation_n=FLAGS.value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )
  elif FLAGS.regret_minimizer == "rm":
    regret_minimizer = regret_matching.RegretMatching(
        mfg_game,
        [],
        eta,
        regret_steps_per_step,
        compress_nus=True,
        compress_every=FLAGS.compress_every,
        compress_lbd=FLAGS.compress_lbd,
        value_estimator=value_estimator,
        value_estimation_n=FLAGS.value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )
  elif FLAGS.regret_minimizer == "poly":
    regret_minimizer = polynomial_weights.PolynomialWeightAlgorithm(
        mfg_game,
        [],
        eta,
        regret_steps_per_step,
        compress_nus=True,
        compress_every=FLAGS.compress_every,
        compress_lbd=FLAGS.compress_lbd,
        value_estimator=value_estimator,
        value_estimation_n=FLAGS.value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )
  else:
    raise NameError(
        "Unknown regret minimizer {}.".format(FLAGS.regret_minimizer)
    )

  if best_responder == "cce":
    best_responder = correlated_equilibrium.cce_br
  elif best_responder == "ce":
    best_responder = correlated_equilibrium.ce_br
  elif best_responder == "ce_partial":
    best_responder = correlated_equilibrium.partial_ce_br
  else:
    raise NameError(
        "Unknown best responder {}. Valid names are `cce` and `ce`.".format(
            FLAGS.best_responder
        )
    )

  mfpsro = mf_psro.MeanFieldPSRO(
      mfg_game,
      regret_minimizer,
      regret_steps_per_step,
      best_responder=best_responder,
  )

  for j in range(FLAGS.n_iter):
    logging.info("Iteration {} of MF-PSRO".format(j))  # pylint: disable=logging-format-interpolation
    print("PSRO Step")
    mfpsro.step()

    print("Equilibrium Computation")
    policies, nus, mus, rhos = mfpsro.get_equilibrium()

    print("Welfare Computation")
    average_welfare = correlated_equilibrium.compute_average_welfare(
        mfg_game, policies, mus, rhos, nus
    )

    print("CCE Gap Computation")
    cce_gap_value = correlated_equilibrium.cce_gap(
        mfg_game, policies, rhos, mus, nus, compute_true_rewards=True
    )
    if compute_ce_gap:
      print("CE Gap Computation")
      ce_gap_value = correlated_equilibrium.ce_gap(
          mfg_game, policies, rhos, mus, nus, compute_true_rewards=True
      )
    else:
      ce_gap_value = 0.0

    print("CCE Gap value : {}".format(cce_gap_value))
    print("CE Gap value : {}".format(ce_gap_value))
    print("Average welfare : {}".format(average_welfare))
    print("")


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Prompts useful for inducing LLM-summarization of debates.
"""

from open_spiel.python.games.chat_games.envs.utils import text


prefix = ('You are an assistant designed to summarize the key arguments in ' +
          'a debate. Please take note of the most import arguments ' +
          'from each side. Provide your summary in 100 ' +
          'words or less. Please summarize the following debate.')
PREFIX = text.wrap([prefix])[0] + '\n\n'

POSTFIX = '\n\nDebate Summary:\n'



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Munchausen deep online mirror descent."""
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import munchausen_deep_mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
import pyspiel


class DeepOnlineMirrorDescentTest(parameterized.TestCase):

  @parameterized.named_parameters(('cpp', 'mfg_crowd_modelling'),
                                  ('python', 'python_mfg_crowd_modelling'))
  def test_train(self, name):
    """Checks that the training works."""
    game = pyspiel.load_game(name)
    assert game.num_players() == 1
    uniform_policy = policy.UniformRandomPolicy(game)
    uniform_dist = distribution.DistributionPolicy(game, uniform_policy)
    env = rl_environment.Environment(
        game, mfg_distribution=uniform_dist, mfg_population=0)
    info_state_size = env.observation_spec()['info_state'][0]
    num_actions = env.action_spec()['num_actions']
    np.random.seed(0)
    args = {
        'alpha': 0.9,
        'batch_size': 128,
        'discount_factor': 1.0,
        'epsilon_decay_duration': 20000000,
        'epsilon_end': 0.1,
        'epsilon_start': 0.1,
        'gradient_clipping': 40,
        'hidden_layers_sizes': [128, 128],
        'learn_every': 64,
        'learning_rate': 0.01,
        'loss': 'mse',
        'min_buffer_size_to_learn': 500,
        'optimizer': 'adam',
        'replay_buffer_capacity': 2000,
        'tau': 10,
        'update_target_network_every': 50
    }
    agent = munchausen_deep_mirror_descent.MunchausenDQN(
        0, info_state_size, num_actions, **args)
    md = munchausen_deep_mirror_descent.DeepOnlineMirrorDescent(
        game, [env], [agent], num_episodes_per_iteration=100)
    for _ in range(10):
      md.iteration()
    nash_conv_md = nash_conv.NashConv(game, md.policy)
    self.assertLessEqual(nash_conv_md.nash_conv(), 3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.stv."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.voting import base
from open_spiel.python.voting import stv


class STVTest(parameterized.TestCase):
  @parameterized.named_parameters(("four", 4), ("one", 1))
  def test_stv_records_number(self, num):
    method = stv.STVVoting(num_winners=num)
    self.assertEqual(
        method.name(), f"single_transferable_vote(num_winners={num})"
    )

  def test_ranked_pairs_wikipedia_example(self):
    alternatives = ["Orange", "Pear", "Strawberry", "Cake", "Chocolate",
                    "Hamburger", "Chicken"]
    votes = [
        base.WeightedVote(4, ["Orange", "Pear"]),
        base.WeightedVote(7, ["Pear", "Strawberry", "Cake"]),
        base.WeightedVote(1, ["Strawberry", "Cake", "Pear"]),
        base.WeightedVote(3, ["Cake", "Chocolate", "Strawberry"]),
        base.WeightedVote(1, ["Cake", "Chocolate", "Hamburger"]),
        base.WeightedVote(4, ["Hamburger"]),
        base.WeightedVote(3, ["Chicken", "Hamburger"]),
    ]
    profile = base.PreferenceProfile(votes=votes,
                                     alternatives=alternatives)
    method = stv.STVVoting(num_winners=3)
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking,
                         ["Pear", "Cake", "Hamburger", "Orange", "Chicken",
                          "Strawberry", "Chocolate"])
    self.assertListEqual(outcome.scores, [14.7, 13.6, 12.7, 7.4, 6.3, 5.2, 4.0])

  def test_meeple_pentathlon(self):
    alternatives = ["A", "B", "C"]
    votes = [
        base.WeightedVote(1, ["A", "B", "C"]),
        base.WeightedVote(1, ["A", "C", "B"]),
        base.WeightedVote(2, ["C", "A", "B"]),
        base.WeightedVote(1, ["B", "C", "A"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = stv.STVVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking, ["C", "A", "B"])
    self.assertListEqual(outcome.scores, [6.3, 3.2, 2.1])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A helpful wrapper for the Elo rating system."""

import pyspiel
from open_spiel.python.voting import base


def compute_ratings_from_preference_profile(
    profile: base.PreferenceProfile,
    smoothing_factor: float = pyspiel.elo.DEFAULT_SMOOTHING_FACTOR,
    max_iterations: int = pyspiel.elo.DEFAULT_MAX_ITERATIONS,
    convergence_delta: float = pyspiel.elo.DEFAULT_CONVERGENCE_DELTA,
) -> dict[str, float]:
  """Compute Elo ratings from a win matrix and a draw matrix."""
  options = pyspiel.elo.default_elo_options()
  options.smoothing_factor = smoothing_factor
  options.max_iterations = max_iterations
  options.convergence_delta = convergence_delta
  match_records = []
  for vote in profile.votes:
    for i in range(len(vote.vote)):
      for j in range(i + 1, len(vote.vote)):
        for _ in range(vote.weight):
          match_records.append(
              pyspiel.elo.MatchRecord(
                  str(vote.vote[i]),
                  str(vote.vote[j]),
                  pyspiel.elo.MatchOutcome.FIRST_PLAYER_WIN,
              )
          )
  return pyspiel.elo.compute_ratings_from_match_records(match_records, options)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model definitions for optimizer network."""

import enum
from typing import Any, Callable, List, Optional, Union

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax


class ModelType(enum.Enum):
  MLP = "MLP"
  RNN = "RNN"


def _mlp_forwards(mlp_hidden_sizes: List[int]) -> hk.Transformed:
  """Returns a haiku transformation of the MLP model to be used in optimizer.

  Args:
    mlp_hidden_sizes: List containing size of linear layers.

  Returns:
    Haiku transformation of the RNN network.
  """
  def forward_fn(inputs):
    mlp = hk.nets.MLP(mlp_hidden_sizes, activation=jax.nn.relu, name="mlp")
    return mlp(inputs)
  return hk.transform(forward_fn)


def _make_rnn_network(lstm_hidden_sizes: List[int],
                      mlp_hidden_sizes: List[int]) -> hk.RNNCore:
  """Returns the RNN network.

  Args:
    lstm_hidden_sizes: List containing size of lstm layers.
    mlp_hidden_sizes: List containing size of linear layers.

  Returns:
    Returns an instance of RNN model.
  """
  layers = []
  for k, hidden_size in enumerate(lstm_hidden_sizes):
    layers += [hk.LSTM(hidden_size, name=f"lstm_layer_{k}"), jax.nn.relu]
  layers += [hk.nets.MLP(mlp_hidden_sizes, name="mlp")]
  return RNNModel(layers)


def _rnn_forwards(lstm_hidden_sizes: List[int], mlp_hidden_sizes: List[int],
                  batch_size: int) -> hk.Transformed:
  """Returns a haiku transformation of the RNN model to be used in optimizer.

  Args:
    lstm_hidden_sizes: List containing size of lstm layers.
    mlp_hidden_sizes: List containing size of linear layers.
    batch_size: Batch size.

  Returns:
    Haiku transformation of the RNN network.
  """
  def forward_fn(inputs):
    rnn = _make_rnn_network(lstm_hidden_sizes, mlp_hidden_sizes)
    initial_state = rnn.initial_state(batch_size=batch_size)
    outputs, _ = hk.dynamic_unroll(rnn, inputs, initial_state, time_major=False)
    return outputs

  return hk.transform(forward_fn)


class RNNModel(hk.RNNCore):
  """RNN model."""

  def __init__(self,
               layers: List[Union[hk.Module, Callable[[jnp.ndarray],
                                                      jnp.ndarray]]],
               name: Optional[str] = None):
    super().__init__(name=name)
    self._layers = layers

  def __call__(self, inputs, prev_state):
    x = inputs
    curr_state = [None] * len(prev_state)
    for k, layer in enumerate(self._layers):
      if isinstance(layer, hk.RNNCore):
        x, curr_state[k] = layer(x, prev_state[k])
      else:
        x = layer(x)
    return x, tuple(curr_state)

  def initial_state(self, batch_size: Optional[int]) -> Any:
    layerwise_init_state = []
    for layer in self._layers:
      if isinstance(layer, hk.RNNCore):
        layerwise_init_state.append(layer.initial_state(batch_size))
      else:
        layerwise_init_state.append(None)
    return tuple(layerwise_init_state)


class OptimizerModel:
  """Optimizer model in l2l paradigm to learn update rules of regret minimizers.

  Attributes:
    mlp_sizes: Size of mlp layers. This is a string, containing sequence of
      numbers, each number indicate size of a linear layer.
    lstm_sizes: Size of lstm layers. This is a string, containing sequence of
      numbers, each number indicate size of an lstm layer.
    initial_learning_rate: Initial value of learning rate used in learning
      rate scheduler.
    batch_size: Batch size.
    num_actions: Number of possible actions.
    num_infostates: Total number of information states.
    model_type: Type of model. For now it can be either MLP or RNN.
    use_infostate_representation: Boolean value to indicate if we use
      information state information as part of model input or not.
    rng: Jax pseudo random number generator.
    model: Neural network model we want to optimize.
    opt_update: Optax optimizer update function.
    net_params: Network parameters.
    opt_state: Optax optimizer state.
    net_apply: Network apply function.
  """

  def __init__(self,
               mlp_sizes: str,
               lstm_sizes: str,
               initial_learning_rate: float,
               batch_size: int,
               num_actions: int,
               num_infostates: int,
               model_type: str = "MLP",
               use_infostate_representation: bool = True):
    self.num_actions = num_actions
    self.num_infostates = num_infostates
    self.initial_learning_rate = initial_learning_rate
    self.batch_size = batch_size
    self.use_infostate_representation = use_infostate_representation
    self.rng = jax.random.PRNGKey(10)

    mlp_sizes_list = [
        int(mlp_size.strip()) for mlp_size in mlp_sizes.split(",")
    ]
    mlp_sizes_list.append(self.num_actions)
    lstm_sizes_list = [
        int(lstm_size.strip()) for lstm_size in lstm_sizes.split(",")
    ]

    if model_type == ModelType.MLP.value:
      self.model = _mlp_forwards(mlp_sizes_list)
    elif model_type == ModelType.RNN.value:
      self.model = _rnn_forwards(lstm_sizes_list, mlp_sizes_list,
                                 self.batch_size)
    else:
      raise ValueError(
          f"{model_type} is not a valid model, model_type should be MLP or RNN."
      )

    self.net_apply = self.model.apply
    self._net_init = self.model.init
    self.opt_update, self.net_params, self.opt_state = None, None, None

  def lr_scheduler(self, init_value: float) -> optax.Schedule:
    schedule_fn = optax.polynomial_schedule(
        init_value=init_value, end_value=0.0001, power=1., transition_steps=100)
    return schedule_fn

  def initialize_optimizer_model(self):
    """Initializes the optax optimizer and neural network model."""
    lr_scheduler_fn = self.lr_scheduler(self.initial_learning_rate)
    opt_init, self.opt_update = optax.chain(
        optax.scale_by_adam(), optax.scale_by_schedule(lr_scheduler_fn),
        optax.scale(-self.initial_learning_rate))

    input_size = self.num_actions
    if self.use_infostate_representation:
      input_size += self.num_infostates

    dummy_input = np.zeros(shape=[self.batch_size, 1, input_size])

    self.net_params = self._net_init(self.rng, dummy_input)
    self.opt_state = opt_init(self.net_params)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A dm_env config for testing a given fixed game with prompt actions.
"""

import collections

from ml_collections import config_dict

from open_spiel.python.games.chat_games.envs.base_envs import email_with_tone
from open_spiel.python.games.chat_games.envs.observations import utils as obs_utils
from open_spiel.python.games.chat_games.envs.payoffs import sentiment
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils
from open_spiel.python.games.chat_games.envs.utils import text as text_utils


def get_config():
  """Get configuration for chat game."""
  config = config_dict.ConfigDict()

  observations = [obs_utils.Observation(),
                  obs_utils.Observation()]

  header = email_with_tone.HEADER

  payoffs = [sentiment.PAYOFF,
             sentiment.PAYOFF]

  given_names = ['Bob',
                 'Suzy']
  num_players = len(given_names)

  given_llm_seeds = [12345]

  given_prompt_actions = collections.OrderedDict()
  tones = ['Happy',
           'Sad',
           'Angry',
           'Calm']
  given_prompt_actions[header.action_keys[0]] = tones
  num_tones = len(tones)

  # Vacuous message
  message = '\n\n'.join(text_utils.wrap(
      ['Hi {receiver},', 'I hope you are well,', 'Best,', '{sender}']
      ))
  initial_scenario = email_with_tone.Scenario(message, 'Bob', 'Suzy', 'Calm')

  query = ('Read the following message. Does it appear that ' +
           'the relevant parties have agreed on a deal? ' +
           'After reading the message, respond Yes or No. ' +
           'Here is the message:\n\n{msg}\n\n')
  llm_termination_prompt = term_utils.Termination(query, '', '')

  params = {'num_distinct_actions': num_players * num_tones,
            'num_llm_seeds': 1,
            'num_init_states': 1,
            'num_players': num_players,
            'min_utility': min([float(p.min) for p in payoffs]),
            'max_utility': max([float(p.max) for p in payoffs]),
            'num_max_replies': 2}

  config.params = params

  config.game = config_dict.ConfigDict()
  config.game.observations = observations
  config.game.header = header
  config.game.payoffs = payoffs
  config.game.given_names = given_names
  config.game.given_llm_seeds = given_llm_seeds
  config.game.given_prompt_actions = given_prompt_actions
  config.game.initial_scenario = initial_scenario
  config.game.llm_list_suffix = 'Output: '
  config.game.llm_termination_prompt = llm_termination_prompt

  return config



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for pairwise data sets."""

import numpy as np
from open_spiel.python.voting import base


def balance_profile(
    profile: base.PreferenceProfile,
    num_votes_per_matchup: int = None,
) -> base.PreferenceProfile:
  """Balances an uneven preference profile.

  Args:
    profile: the preference profile to balance.
    num_votes_per_matchup: the number of votes to cast for each matchup. If
      None, then the function will take the maximum count of any matchup in the
      profile and use that.

  Returns:
    A new preference profile with balanced pairwise counts.

  An uneven profile is one where the distribution over number of matchups
  between (i, j) is non-uniform. This function creates a new profile
  (of pairwise preferences) where the number of votes per (i, j) matchup is
  uniform, while (in expectation) maintaining the true data's win rates per
  matchup.
  """
  m = profile.num_alternatives()
  new_profile = base.PreferenceProfile(alternatives=profile.alternatives)

  profile_counts = profile.pairwise_count_matrix()
  pref_matrix = profile.pref_matrix().astype(float)

  if num_votes_per_matchup is None:
    num_votes_per_matchup = profile.pairwise_count_matrix().max()

  print(f"num_votes_per_matchup: {num_votes_per_matchup}")

  # Add I to the counts to prevent the divide-by-zero on the diagonal.
  winrate_mat = pref_matrix.astype(float) / (profile_counts + np.eye(m))
  assert profile.alternatives == new_profile.alternatives

  for i in range(m):
    for j in range(i+1, m):
      alt_i = new_profile.alternatives[i]
      alt_j = new_profile.alternatives[j]
      num_wins = int(round(winrate_mat[i, j] * num_votes_per_matchup))
      num_losses = max(0, num_votes_per_matchup - num_wins)
      for _ in range(num_wins):
        new_profile.add_vote([alt_i, alt_j])
      for _ in range(num_losses):
        new_profile.add_vote([alt_j, alt_i])

  return new_profile




# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Kemeny-Young method.

Based on https://en.wikipedia.org/wiki/Kemeny%E2%80%93Young_method.
"""

import itertools
from typing import List, Tuple
import numpy as np
from open_spiel.python.voting import base


class KemenyYoungVoting(base.AbstractVotingMethod):
  """Implements Kemeny-Young's method."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "kemeny_young"

  def _score(
      self,
      pref_mat: np.ndarray,
      perm: Tuple[int, ...],
  ) -> np.ndarray:
    # The score of alternative a_i in a ranking R is defined to be:
    #      KemenyScore(a_i) = sum_{a_j s.t. R(a_i) >= R(a_j)} N(a_i, a_j)
    # The score of ranking R is then sum_i KemenyScore(a_i).
    num_alts = len(perm)
    scores = np.zeros(num_alts, dtype=np.int32)
    for i in range(num_alts):
      for j in range(i+1, num_alts):
        scores[i] += pref_mat[perm[i], perm[j]]
    return scores

  def _permutation_to_ranking(
      self,
      alternatives: List[base.AlternativeId],
      permutation: Tuple[base.AlternativeId, ...]) -> List[base.AlternativeId]:
    assert len(permutation) == len(alternatives)
    return [alternatives[permutation[i]] for i in range(len(alternatives))]

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    pref_mat = profile.pref_matrix()
    alternatives = profile.alternatives
    m = profile.num_alternatives()
    best_permutation = None
    best_score = -1
    best_score_array = None
    for permutation in itertools.permutations(range(m)):
      scores = self._score(pref_mat, permutation)
      total_score = scores.sum()
      if total_score > best_score:
        best_score = total_score
        best_score_array = scores
        best_permutation = permutation
    best_ranking = self._permutation_to_ranking(alternatives, best_permutation)
    outcome = base.RankOutcome(rankings=best_ranking,
                               scores=list(best_score_array))
    return outcome




# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pyspiel config for meta-generated real-world negotiation games.
"""

import collections

from ml_collections import config_dict

from open_spiel.python.games.chat_games.envs.base_envs import email_with_tone
from open_spiel.python.games.chat_games.envs.observations import summary
from open_spiel.python.games.chat_games.envs.observations import utils as obs_utils
from open_spiel.python.games.chat_games.envs.payoffs import sentiment
from open_spiel.python.games.chat_games.envs.scenarios.actions import tones
from open_spiel.python.games.chat_games.envs.scenarios.domains import real_world_negotiations as rwn
from open_spiel.python.games.chat_games.envs.scenarios.players import names
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils


def get_config():
  """Get configuration for chat game."""
  config = config_dict.ConfigDict()

  num_players = 3

  observations = [
      obs_utils.Observation(summary.PREFIX, summary.POSTFIX)
      for _ in range(num_players)
  ]

  scenario_a = email_with_tone.Scenario(rwn.SCENARIO_A, 'Alice', 'Bob')
  scenario_b = email_with_tone.Scenario(rwn.SCENARIO_B, 'Joel', 'Gene')
  scenario_c = email_with_tone.Scenario(rwn.SCENARIO_C, 'George', 'Jill')
  examples_scenarios = [scenario_a,
                        scenario_b,
                        scenario_c]

  header = email_with_tone.HEADER

  payoffs = [sentiment.PAYOFF]

  examples_names = names.NAMES

  examples_prompt_actions = collections.OrderedDict()
  examples_prompt_actions[header.action_keys[0]] = tones.TONES
  num_tones = 3

  query = ('Read the following message. Does it appear that ' +
           'the relevant parties have agreed on a deal? ' +
           'After reading the message, respond Yes or No. ' +
           'Here is the message:\n\n{msg}\n\n')
  llm_termination_prompt = term_utils.Termination(query, '', '')

  params = {'num_distinct_actions': num_players * num_tones,
            'num_llm_seeds': 2,
            'num_init_states': 3,
            'num_players': num_players,
            'min_utility': min([float(p.min) for p in payoffs]),
            'max_utility': max([float(p.max) for p in payoffs]),
            'num_max_replies': 2}

  config.params = params

  config.game = config_dict.ConfigDict()
  config.game.observations = observations
  config.game.header = header
  config.game.payoffs = payoffs
  config.game.num_names = 10
  config.game.num_prompt_actions = (num_tones,)
  config.game.num_private_info = (3,)
  config.game.examples_names = examples_names
  config.game.examples_prompt_actions = examples_prompt_actions
  config.game.examples_scenarios = examples_scenarios
  config.game.llm_list_suffix = 'Output: '
  config.game.llm_termination_prompt = llm_termination_prompt

  return config



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This compares the speed/results of Jax CFR to of the original impl of CFR.

The results slightly differ due to different rounding of regrets between
original implmentation and CFR. When setting clamping of regrets to 1e-8 the
results are exactly the same.
"""


# pylint: disable=g-importing-member

import time
from open_spiel.python.algorithms.best_response import BestResponsePolicy
from open_spiel.python.algorithms.cfr import CFRPlusSolver
from open_spiel.python.jax.cfr.jax_cfr import JaxCFR
import pyspiel


def compare_cfr_with_jax_cfr(game):
  """Do the comparison."""

  start = time.time()
  jax_cfr = JaxCFR(game)
  print(time.time() - start)
  jax_cfr.multiple_steps(10000)
  print(time.time() - start)

  start = time.time()
  print(time.time() - start)
  cfr = CFRPlusSolver(game)
  for _ in range(1000):
    cfr.evaluate_and_update_policy()

  print(time.time() - start)

  jax_strat = jax_cfr.average_policy()
  jax_br1 = BestResponsePolicy(jax_cfr.game, 1, jax_strat)
  jax_br2 = BestResponsePolicy(jax_cfr.game, 0, jax_strat)

  cfr_strat = jax_cfr.average_policy()
  cfr_br1 = BestResponsePolicy(jax_cfr.game, 1, cfr_strat)
  cfr_br2 = BestResponsePolicy(jax_cfr.game, 0, cfr_strat)

  print("Jax P1: ", jax_br1.value(jax_cfr.game.new_initial_state()))
  print("CFR P1: ", cfr_br1.value(jax_cfr.game.new_initial_state()))
  print("Jax P2: ", jax_br2.value(jax_cfr.game.new_initial_state()))
  print("CFR P2: ", cfr_br2.value(jax_cfr.game.new_initial_state()))


# Speed Results:
#   Original: 139.60753107070923
#   Jax CPU: 3.7404067516326904
def compare_leduc():
  game = pyspiel.load_game("leduc_poker")
  compare_cfr_with_jax_cfr(game)


# Speed Results:
#   Original: 335.6707363128662
#   Jax CPU: 7.59996485710144
def compare_battleship():
  game_params = {
      "board_height": 2,
      "board_width": 2,
      "num_shots": 4,
      "ship_sizes": "[2]",
      "ship_values": "[1]",
      "allow_repeated_shots": False,
  }
  game = pyspiel.load_game("battleship", game_params)
  compare_cfr_with_jax_cfr(game)


# Speed Results:
#   Original: 14.667663097381592
#   Jax CPU: 1.068636417388916
def compare_goofspiel_descending():
  game_params = {"num_cards": 4, "imp_info": True, "points_order": "descending"}
  game = pyspiel.load_game_as_turn_based("goofspiel", game_params)
  compare_cfr_with_jax_cfr(game)


# Speed Results:
#   Original: 6.639796733856201
#   Jax CPU: 0.8599820137023926
def compare_goofspiel_randomized():
  game_params = {"num_cards": 3, "imp_info": True, "points_order": "random"}
  game = pyspiel.load_game_as_turn_based("goofspiel", game_params)
  compare_cfr_with_jax_cfr(game)


if __name__ == "__main__":
  compare_leduc()
  compare_battleship()
  compare_goofspiel_descending()
  compare_goofspiel_randomized()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.policy_aggregator."""

import unittest
from absl.testing import parameterized

import numpy as np

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import policy_aggregator
import pyspiel


class PolicyAggregatorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          "testcase_name": "kuhn_poker",
          "game_name": "kuhn_poker"
      }, {
          "testcase_name": "leduc_poker",
          "game_name": "leduc_poker"
      })
  def test_policy_aggregation_random(self, game_name):
    env = rl_environment.Environment(game_name)

    policies = [[policy.UniformRandomPolicy(env.game)
                 for _ in range(2)]
                for _ in range(2)]
    probabilities = [
        list(np.ones(len(policies)) / len(policies)) for _ in range(2)
    ]

    pol_ag = policy_aggregator.PolicyAggregator(env.game)
    aggr_policy = pol_ag.aggregate([0], policies, probabilities)

    for item in aggr_policy.policy[0].items():
      _, probs = zip(*item[1].items())
      const_probs = tuple([probs[0]] * len(probs))
      self.assertEqual(probs, const_probs)

  @parameterized.named_parameters(
      {
          "testcase_name": "kuhn_poker",
          "game_name": "kuhn_poker"
      }, {
          "testcase_name": "leduc_poker",
          "game_name": "leduc_poker"
      })
  def test_policy_aggregation_tabular_randinit(self, game_name):
    env = rl_environment.Environment(game_name)

    mother_policy = policy.TabularPolicy(env.game).copy_with_noise(
        1, 10, np.random.RandomState(0))
    policies = [[mother_policy.__copy__() for _ in range(2)] for _ in range(2)]
    probabilities = [
        list(np.ones(len(policies)) / len(policies)) for _ in range(2)
    ]

    pol_ag = policy_aggregator.PolicyAggregator(env.game)
    aggr_policy = pol_ag.aggregate([0], policies, probabilities)

    for state, value in aggr_policy.policy[0].items():
      polici = mother_policy.policy_for_key(state)

      value_normal = {
          action: probability
          for action, probability in enumerate(polici)
          if probability > 0
      }
      for key in value_normal.keys():
        self.assertAlmostEqual(value[key], value_normal[key], 8)

  @parameterized.named_parameters({
      "testcase_name": "tic_tac_toe",
      "game_name": "tic_tac_toe",
  })
  def test_policy_aggregation_variadic(self, game_name):
    game = pyspiel.load_game(game_name)

    uniform_policy = policy.UniformRandomPolicy(game)
    first_action_policy = policy.FirstActionPolicy(game)

    pol_ag = policy_aggregator.PolicyAggregator(game)

    weights0 = [1.0, 0.0]
    player0 = pol_ag.aggregate(
        list(range(game.num_players())),
        [[uniform_policy, first_action_policy]] + [[uniform_policy]] *
        (game.num_players() - 1),
        [weights0] + [[1.0]] * (game.num_players() - 1))
    state = game.new_initial_state()
    action_prob = player0.action_probabilities(state)
    for action in action_prob:
      if action_prob[action] > 0:
        self.assertAlmostEqual(action_prob[action],
                               1. / len(state.legal_actions()))


if __name__ == "__main__":
  unittest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean-Field Correlated Equilibrium Gap & Best Response Computation Library.

"""

import numpy as np
from open_spiel.python.mfg.algorithms import greedy_policy
from open_spiel.python.mfg.algorithms import joint_best_response_value as jbr
from open_spiel.python.mfg.algorithms import utils


def get_joint_br(game, weights, mus):
  br_value = jbr.JointBestResponse(game, mus, weights)
  greedy_pi = greedy_policy.GreedyPolicy(game, None, br_value)
  return greedy_pi, br_value


def compute_rewards(game, policies, mus):
  return np.array([
      [utils.get_exact_value(pi, mu, game) for pi in policies] for mu in mus
  ])


def compute_average_welfare(game, policies, mus, rhos, nus):
  """Computes average welfare.

  Args:
    game: Pyspiel game.
    policies: List of policies, length P
    mus: List of State Distributions of length T
    rhos: Temporal weights, length T
    nus: Policy distribution per time, shape [T, P]

  Returns:
    Average welfare.
  """
  assert len(mus) == len(rhos)
  assert len(rhos) == nus.shape[0]
  assert len(policies) == nus.shape[1]

  rewards = compute_rewards(game, policies, mus)
  return np.sum(rewards * nus * rhos.reshape(-1, 1))


def cce_br(game, policies, weights, mus, nus, rewards=None):
  """Computes CCE-BR.

  Args:
    game: Pyspiel MFG Game.
    policies: List of pyspiel policies, length P.
    weights: Array of temporal weights on each distribution in `nu`, length T.
    mus: List of state distributions, length T.
    nus: Array of policy distribution per timestep, shape (T, P)
    rewards: Optional array of policy reward per timestep, shape (T, P)

  Returns:
    Best-response, computed exploitability from `rewards`.
  """
  assert len(mus) == len(nus)
  assert len(mus) == len(weights)

  del policies
  pol, val = get_joint_br(game, weights, mus)
  cce_gap_value = None
  if len(rewards) > 0:  # pylint: disable=g-explicit-length-test
    deviation_value = val.value(game.new_initial_states()[0])
    on_policy_value = np.sum(weights * np.sum(rewards * nus, axis=1))
    cce_gap_value = deviation_value - on_policy_value
  return [pol], cce_gap_value


def ce_br(game, policies, weights, mus, nus, rewards=None):
  """Computes CE-BR.

  Args:
    game: Pyspiel MFG Game.
    policies: List of pyspiel policies, length P.
    weights: Array of temporal weights on each distribution in `nu`, length T.
    mus: List of state distributions, length T.
    nus: Array of policy distribution per timestep, shape (T, P)
    rewards: Optional array of policy reward per timestep, shape (T, P)

  Returns:
    Best-responses, computed exploitability from `rewards`.
  """
  assert len(mus) == len(nus)
  assert len(mus) == len(weights)

  policy_probability = np.sum(nus, axis=0)
  new_policies = []
  ce_gap_value = 0.0
  nus = np.array(nus)
  weights = np.array(weights)
  for policy_index in range(len(policies)):
    if policy_probability[policy_index] > 0:
      # Take conditional distribution
      pol_weights = nus[:, policy_index] * weights
      pol_proba = np.sum(pol_weights)
      pol_weights = pol_weights / pol_proba

      # Prune state distribution and weights from 0.0-weightred values
      new_mus = [mu for ind, mu in enumerate(mus) if pol_weights[ind] > 0]
      new_weights = np.array([
          weight for ind, weight in enumerate(pol_weights)
          if pol_weights[ind] > 0
      ])

      # Compute best-response.
      new_pol, new_val = get_joint_br(game, new_weights, new_mus)
      new_br_val = new_val.value(game.new_initial_states()[0])

      # Evaluate CE-Gap
      if len(rewards) > 0:  # pylint: disable=g-explicit-length-test
        on_policy_value = np.sum(
            np.array(rewards)[:, policy_index] * pol_weights)
        ce_gap_value += pol_proba * (new_br_val - on_policy_value)
      new_policies.append(new_pol)
  return new_policies, ce_gap_value


def partial_ce_br(game, policies, weights, mus, nus, rewards=None):
  """Computes CE-BR for a single sampled policy.

  Args:
    game: Pyspiel MFG Game.
    policies: List of pyspiel policies, length P.
    weights: Array of temporal weights on each distribution in `nu`, length T.
    mus: List of state distributions, length T.
    nus: Array of policy distribution per timestep, shape (T, P)
    rewards: Optional array of policy reward per timestep, shape (T, P)

  Returns:
    Best-response, noisy exploitability estimation.
  """
  policy_probability = np.sum(nus, axis=0)
  new_policies = []

  ce_gap_value = None
  policy_index = np.random.choice(list(range(len(policies))))
  if policy_probability[policy_index] > 0:
    # Take conditional distribution
    pol_weights = [nu[policy_index] * weight for nu, weight in zip(
        nus, weights)]
    pol_proba = np.sum(pol_weights)
    pol_weights = np.array(pol_weights) / pol_proba

    # Prune state distribution and weights from 0.0-weightred values
    new_mus = [mu for ind, mu in enumerate(mus) if pol_weights[ind] > 0]
    new_weights = [
        weight for ind, weight in enumerate(pol_weights)
        if pol_weights[ind] > 0
    ]

    # Compute best-response.
    new_pol, new_val = get_joint_br(game, new_weights, new_mus)
    new_br_val = new_val.value(game.new_initial_states()[0])

    # Evaluate CE-Gap
    if len(rewards) > 0:  # pylint: disable=g-explicit-length-test
      on_policy_value = np.sum(np.array(rewards)[:, policy_index] * pol_weights)
      ce_gap_value = (new_br_val - on_policy_value)
    new_policies.append(new_pol)
  return new_policies, ce_gap_value


def cce_gap(game, policies, weights, mus, nus, rewards=None,
            compute_true_rewards=False):
  if compute_true_rewards:
    rewards = compute_rewards(game, policies, mus)
  assert rewards is not None, ("Must provide rewards matrix when computing CCE "
                               "Gap.")
  _, gap = cce_br(game, policies, weights, mus, nus, rewards=rewards)
  return gap


def ce_gap(game, policies, weights, mus, nus, rewards=None,
           compute_true_rewards=False):
  if compute_true_rewards:
    rewards = compute_rewards(game, policies, mus)
  assert rewards is not None, ("Must provide rewards matrix when computing CE "
                               "Gap.")
  _, gap = ce_br(game, policies, weights, mus, nus, rewards=rewards)
  return gap



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meta-strategy solvers for PSRO."""

import numpy as np

from open_spiel.python.algorithms import lp_solver
from open_spiel.python.algorithms import projected_replicator_dynamics
from open_spiel.python.algorithms import regret_matching
import pyspiel


EPSILON_MIN_POSITIVE_PROBA = 1e-8


def uniform_strategy(solver, return_joint=False):
  """Returns a Random Uniform distribution on policies.

  Args:
    solver: GenPSROSolver instance.
    return_joint: If true, only returns marginals. Otherwise marginals as well
      as joint probabilities.

  Returns:
    uniform distribution on strategies.
  """
  policies = solver.get_policies()
  policy_lengths = [len(pol) for pol in policies]
  result = [np.ones(pol_len) / pol_len for pol_len in policy_lengths]
  if not return_joint:
    return result
  else:
    joint_strategies = get_joint_strategy_from_marginals(result)
    return result, joint_strategies


def softmax_on_range(number_policies):
  x = np.array(list(range(number_policies)))
  x = np.exp(x-x.max())
  x /= np.sum(x)
  return x


def uniform_biased_strategy(solver, return_joint=False):
  """Returns a Biased Random Uniform distribution on policies.

  The uniform distribution is biased to prioritize playing against more recent
  policies (Policies that were appended to the policy list later in training)
  instead of older ones.

  Args:
    solver: GenPSROSolver instance.
    return_joint: If true, only returns marginals. Otherwise marginals as well
      as joint probabilities.

  Returns:
    uniform distribution on strategies.
  """
  policies = solver.get_policies()
  if not isinstance(policies[0], list):
    policies = [policies]
  policy_lengths = [len(pol) for pol in policies]
  result = [softmax_on_range(pol_len) for pol_len in policy_lengths]
  if not return_joint:
    return result
  else:
    joint_strategies = get_joint_strategy_from_marginals(result)
    return result, joint_strategies


def renormalize(probabilities):
  """Replaces all negative entries with zeroes and normalizes the result.

  Args:
    probabilities: probability vector to renormalize. Has to be one-dimensional.

  Returns:
    Renormalized probabilities.
  """
  probabilities[probabilities < 0] = 0
  probabilities = probabilities / np.sum(probabilities)
  return probabilities


def get_joint_strategy_from_marginals(probabilities):
  """Returns a joint strategy tensor from a list of marginals.

  Args:
    probabilities: list of probabilities.

  Returns:
    A flat joint strategy from a list of marginals.
  """
  res = np.ones((1,), dtype=np.float64)
  for prob in probabilities:
    res = res[..., None] @ np.asarray(prob).reshape((1,) * res.ndim + (-1,))
  return res.reshape(-1)


def nash_strategy(solver, return_joint=False):
  """Returns nash distribution on meta game matrix.

  This method only works for two player zero-sum games.

  Args:
    solver: GenPSROSolver instance.
    return_joint: If true, only returns marginals. Otherwise marginals as well
      as joint probabilities.

  Returns:
    Nash distribution on strategies.
  """
  meta_games = solver.get_meta_game()
  if not isinstance(meta_games, list):
    meta_games = [meta_games, -meta_games]
  meta_games = [x.tolist() for x in meta_games]
  if len(meta_games) != 2:
    raise NotImplementedError(
        "nash_strategy solver works only for 2p zero-sum"
        "games, but was invoked for a {} player game".format(len(meta_games)))
  nash_prob_1, nash_prob_2, _, _ = (
      lp_solver.solve_zero_sum_matrix_game(
          pyspiel.create_matrix_game(*meta_games)))
  result = [
      renormalize(np.array(nash_prob_1).reshape(-1)),
      renormalize(np.array(nash_prob_2).reshape(-1))
  ]

  if not return_joint:
    return result
  else:
    joint_strategies = get_joint_strategy_from_marginals(result)
    return result, joint_strategies


def prd_strategy(solver, return_joint=False):
  """Computes Projected Replicator Dynamics strategies.

  Args:
    solver: GenPSROSolver instance.
    return_joint: If true, only returns marginals. Otherwise marginals as well
      as joint probabilities.

  Returns:
    PRD-computed strategies.
  """
  meta_games = solver.get_meta_game()
  if not isinstance(meta_games, list):
    meta_games = [meta_games, -meta_games]
  kwargs = solver.get_kwargs()
  result = projected_replicator_dynamics.projected_replicator_dynamics(
      meta_games, **kwargs)
  if not return_joint:
    return result
  else:
    joint_strategies = get_joint_strategy_from_marginals(result)
    return result, joint_strategies


def rm_strategy(solver, return_joint=False):
  """Computes regret-matching strategies.

  Args:
    solver: GenPSROSolver instance.
    return_joint: If true, only returns marginals. Otherwise marginals as well
      as joint probabilities.

  Returns:
    PRD-computed strategies.
  """
  meta_games = solver.get_meta_game()
  if not isinstance(meta_games, list):
    meta_games = [meta_games, -meta_games]
  kwargs = solver.get_kwargs()
  result = regret_matching.regret_matching(meta_games, **kwargs)
  if not return_joint:
    return result
  else:
    joint_strategies = get_joint_strategy_from_marginals(result)
    return result, joint_strategies


META_STRATEGY_METHODS = {
    "uniform_biased": uniform_biased_strategy,
    "uniform": uniform_strategy,
    "nash": nash_strategy,
    "prd": prd_strategy,
    "rm": rm_strategy,
}



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some general utility functions for coalitional games."""

import itertools
import numpy as np
from open_spiel.python.coalitional_games import coalitional_game


def compute_payoff_epsilon(
    game: coalitional_game.CoalitionalGame,
    p: np.ndarray
) -> float:
  """For a payoff vector p, get max_e s.t. p dot c + e >= V(c).

  Warning! Enumerates all coalitions.

  Args:
    game: the game to enumerate.
    p: the payoff vector.

  Returns:
    the value max_e s.t. p dot c + e >= V(C) for all subsets C subseteq N.
  """
  epsilon = 0
  for c in itertools.product([0, 1], repeat=game.num_players()):
    coalition = np.asarray(c)
    val_c = game.coalition_values(coalition)
    payoffs_to_coalition = np.inner(p, coalition)
    epsilon = max(epsilon, val_c - payoffs_to_coalition)
  return epsilon



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Does a backward pass to output the value of a policy."""
from typing import Optional
from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import distribution as distribution_std
from open_spiel.python.mfg import value
import pyspiel


class PolicyValue(value.ValueFunction):
  """Computes the value of a specified strategy."""

  def __init__(self,
               game,
               distribution: distribution_std.Distribution,
               policy: policy_std.Policy,
               state_value: Optional[value.ValueFunction] = None,
               root_state=None):
    """Initializes the value calculation.

    Args:
      game: The game to analyze.
      distribution: A `distribution.Distribution` object.
      policy: A `policy.Policy` object.
      state_value: A state value function. Defaults to Tabular.
      root_state: The state of the game at which to start. If `None`, the game
        root state is used.
    """
    super(PolicyValue, self).__init__(game)
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._distribution = distribution
    self._policy = policy

    self._state_value = (state_value if state_value is not None
                         else value.TabularValueFunction(game))

    self.evaluate()

  def eval_state(self, state):
    """Evaluate the value of a state."""
    state_str = state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    if self._state_value.has(state_str):
      return self._state_value(state_str)
    elif state.is_terminal():
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()])
      return self._state_value(state_str)
    elif state.current_player() == pyspiel.PlayerId.CHANCE:
      self._state_value.set_value(state_str, 0.0)
      for action, prob in state.chance_outcomes():
        new_state = state.child(action)
        self._state_value.add_value(state_str,
                                    prob * self.eval_state(new_state))
      return self._state_value(state_str)
    elif state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
      dist_to_register = state.distribution_support()
      dist = [
          self._distribution.value_str(str_state, 0.)
          for str_state in dist_to_register
      ]
      new_state = state.clone()
      new_state.update_distribution(dist)
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()] +
          self.eval_state(new_state))
      return self._state_value(state_str)
    else:
      assert int(state.current_player()) >= 0, "The player id should be >= 0"
      v = 0.0
      for action, prob in self._policy.action_probabilities(state).items():
        new_state = state.child(action)
        v += prob * self.eval_state(new_state)
      self._state_value.set_value(
          state_str,
          state.rewards()[state.mean_field_population()] + v)
      return self._state_value(state_str)

  def evaluate(self):
    """Evaluate the value over states of self._policy."""
    for state in self._root_states:
      self.eval_state(state)

  def value(self, state, action=None):
    if action is None:
      return self._state_value(
          state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))
    new_state = state.child(action)
    return state.rewards()[state.mean_field_population()] + self._state_value(
        new_state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Typing definitions."""

from typing import Any, Dict, Callable
import jax.numpy as jnp
import optax
from open_spiel.python.examples.meta_cfr.sequential_games import game_tree_utils

PyTree = Any
Params = PyTree
ApplyFn = Callable[..., jnp.ndarray]
OptState = optax.OptState

GameTree = game_tree_utils.GameTree
InfostateNode = game_tree_utils.InfoState
InfostateMapping = Dict[str, InfostateNode]
HistoryNode = game_tree_utils.HistoryTreeNode




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for chess."""


from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

import pyspiel

blackjack = pyspiel.blackjack

NUM_SIM_GAMES = 10
SEED = 87375711


class GamesBlackjackTest(parameterized.TestCase):

  def test_blackjack_game_sim(self):
    game = pyspiel.load_game("blackjack")
    for _ in range(NUM_SIM_GAMES):
      print("----------------")
      print("New game")
      print("----------------")
      state = game.new_initial_state()
      while not state.is_terminal():
        if state.is_chance_node():
          outcomes = state.chance_outcomes()
          action_list, prob_list = zip(*outcomes)
          action = np.random.choice(action_list, p=prob_list)
          print("Chance samples", state.action_to_string(action))
        else:
          print("Player turn")
          print("My cards:", blackjack.cards_to_strings(state.cards(0)))
          print("My best total:", state.get_best_player_total(0))
          print(
              "Dealer's cards:",
              blackjack.cards_to_strings(state.cards(state.dealer_id())),
          )
          print(
              "Dealer's best total:",
              state.get_best_player_total(state.dealer_id()),
          )
          actions = state.legal_actions()
          action = np.random.choice(actions)
          print("Action chosen:", state.action_to_string(action))
        state.apply_action(action)
        print("")
      print("Terminal state: ")
      print(str(state))
      print("Returns:", state.returns())
      print("")

  def test_card_to_string_conversion(self):
    for i in range(52):
      self.assertEqual(i, blackjack.get_card_by_string(
          blackjack.card_to_string(i)))

  def test_blackjack_three_aces(self):
    game = pyspiel.load_game("blackjack")
    state = game.new_initial_state()
    self.assertTrue(state.is_chance_node())
    # Player's cards
    state.apply_action(blackjack.get_card_by_string("D6"))
    state.apply_action(blackjack.get_card_by_string("DA"))
    # Dealer's cards
    state.apply_action(blackjack.get_card_by_string("CQ"))
    state.apply_action(blackjack.get_card_by_string("C3"))
    # Play starts.
    self.assertFalse(state.is_chance_node())
    self.assertListEqual(blackjack.cards_to_strings(state.cards(0)),
                         ["D6", "DA"])
    self.assertEqual(state.get_best_player_total(0), 17)
    state.apply_action(blackjack.HIT)
    state.apply_action(blackjack.get_card_by_string("SA"))
    self.assertEqual(state.get_best_player_total(0), 18)
    self.assertListEqual(blackjack.cards_to_strings(state.cards(0)),
                         ["D6", "DA", "SA"])
    state.apply_action(blackjack.HIT)
    state.apply_action(blackjack.get_card_by_string("CA"))
    self.assertEqual(state.get_best_player_total(0), 19)
    self.assertListEqual(blackjack.cards_to_strings(state.cards(0)),
                         ["D6", "DA", "SA", "CA"])
    state.apply_action(blackjack.HIT)
    state.apply_action(blackjack.get_card_by_string("C2"))
    self.assertEqual(state.get_best_player_total(0), 21)
    state.apply_action(blackjack.STAND)
    self.assertListEqual(blackjack.cards_to_strings(state.cards(0)),
                         ["D6", "DA", "SA", "CA", "C2"])
    # Dealer's turn.
    # Dealer has a 13, must hit.
    self.assertTrue(state.is_chance_node())
    self.assertListEqual(
        blackjack.cards_to_strings(state.cards(state.dealer_id())),
        ["CQ", "C3"])
    state.apply_action(blackjack.get_card_by_string("HA"))
    self.assertEqual(state.get_best_player_total(1), 14)
    self.assertTrue(state.is_chance_node())
    self.assertListEqual(
        blackjack.cards_to_strings(state.cards(state.dealer_id())),
        ["CQ", "C3", "HA"])
    state.apply_action(blackjack.get_card_by_string("S3"))
    self.assertEqual(state.get_best_player_total(1), 17)
    self.assertListEqual(
        blackjack.cards_to_strings(state.cards(state.dealer_id())),
        ["CQ", "C3", "HA", "S3"])
    # Dealer must stop on 17. This should be a terminal state.
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns(), [1.0])


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.approval."""

from absl.testing import absltest

from open_spiel.python.voting import approval
from open_spiel.python.voting import base


class ApprovalVotingTest(absltest.TestCase):

  def test_approval_name_correct(self):
    method = approval.ApprovalVoting(k=7)
    self.assertEqual(method.name(), "approval(k=7)")

  def test_approval_basic_run(self):
    votes = [
        ["a", "b", "c", "d"],
        ["b", "d", "a", "c"],
        ["a", "c", "d", "b"],
        ["d", "b", "c", "a"]
    ]
    profile = base.PreferenceProfile(votes=votes)
    method = approval.ApprovalVoting(k=2)
    outcome = method.run_election(profile)
    with self.subTest("Approval voting gets basic ranking correct"):
      self.assertTrue(outcome.ranking == ["b", "d", "a", "c"] or
                      outcome.ranking == ["b", "a", "d", "c"])
    with self.subTest("Approval voting gets basic scores correct"):
      self.assertListEqual(outcome.scores, [3, 2, 2, 1])

  def test_approval_basic_run_with_weights(self):
    votes = [
        base.WeightedVote(1, ["a", "b", "c", "d"]),
        base.WeightedVote(2, ["b", "d", "a", "c"]),
        base.WeightedVote(3, ["a", "c", "d", "b"]),
        base.WeightedVote(4, ["d", "b", "c", "a"])
    ]
    profile = base.PreferenceProfile(votes=votes)
    method = approval.ApprovalVoting(k=2)
    outcome = method.run_election(profile)
    with self.subTest("Approval voting gets weighted ranking correct"):
      self.assertListEqual(outcome.ranking, ["b", "d", "a", "c"])
    with self.subTest("Approval voting gets weighted scores correct"):
      self.assertListEqual(outcome.scores, [7, 6, 4, 3])


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plurality voting method.

Based on https://en.wikipedia.org/wiki/Plurality_voting.
"""

from open_spiel.python.voting import base


class PluralityVoting(base.AbstractVotingMethod):
  """Implements the plurality (first past the post) voting rule."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "plurality"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    tally = {}
    for alternative in profile.alternatives:
      tally[alternative] = 0
    for vote in profile.votes:
      tally[vote.vote[0]] += vote.weight
    sorted_tally = sorted(tally.items(), key=lambda item: item[1], reverse=True)
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_tally)
    return outcome
